/* 
 * File:   Dotplot_Wrapper.cc
 * Author: TCO
 * 
 * Created on 30. September 2009, 11:34
 */

#include "Dotplot_Wrapper.hh"

Dotplot_Wrapper::Dotplot_Wrapper(std::string fname) {
    fileHandle_.open(fname.c_str());
    seq_id_ = seqname_from_filename(fname);

    if (!fileHandle_.good()){
        std::cerr << "Cannot read "<< fname << std::endl;
	std::exit(-1);
    }else{
        fetch_sequence();
        ready = prepare_reading(); // search for the basepairs location
    }

}

std::string Dotplot_Wrapper::get_seqID(){
    return std::string(seq_id_);
}

std::string Dotplot_Wrapper::get_sequence(){
    return std::string(sequence_);
}

std::string Dotplot_Wrapper::seqname_from_filename(const std::string &s) {
	score_t i;
	score_t j;

        // Search backwards for the last path separator
	for (i=s.length()-1; i>=0 && s[i]!='/'; i--)
		;

        // Search forward for the dot of the file suffix
	for (j=i+1; j<(score_t)s.length() && s[j]!='.'; j++)
		;

	std::string name=s.substr(i+1, j-i-1);   // name is between separator and dot

	if (name.length()>3 && name.substr(name.length()-3, 3) == "_dp") // trunc "_dp" if present
		name=name.substr(0, name.length()-3);

	return name;
}

void Dotplot_Wrapper::fetch_sequence(){
        fileHandle_.seekg(std::ios_base::beg); // Just in case assure that we read the file from beginning
        std::string tmp_seq;            // Search for the right position labeled "sequence"
	while (!fileHandle_.eof() && fileHandle_ >> tmp_seq && tmp_seq.compare("/sequence") != 0);

        sequence_.clear();

	if (fileHandle_.eof()){
            std::cerr << "Problems reading the dotplot!" << std::endl;
	    std::exit(-1);
	}

	char c;
        tmp_seq.clear();

	// skip the following two braces
        while (!fileHandle_.eof() && fileHandle_ >> c && c != '{');
	while (!fileHandle_.eof() && fileHandle_ >> c && c != '(');
        // read...
        while (!fileHandle_.eof() && fileHandle_ >> c && c != ')')
		if ((c != '\\') && (c != '\n'))
                    sequence_ += c;

}


bool Dotplot_Wrapper::has_more_basepairs(){
    return ready;
}

/*
 * Search for the basepair position in the dotplot and return the success
 * Also read ne NEXT basepair.
 *
 */
bool Dotplot_Wrapper::prepare_reading(){
    fileHandle_.seekg(std::ios_base::beg);
    std::string buf;
    while (std::getline(fileHandle_, buf) && buf != "%data starts here" && buf !=  "drawgrid_turn" ) ;

    if (fileHandle_.eof()){
        std::cerr << "Problems reading the dotplot!" << std::endl;
        return false;
    }
    else{
        ready = true;
        read_basepair();    // read the next basepair. The current value will be discarded. Should be uninitialized anyway
    }
    return ready;
}

basepair Dotplot_Wrapper::read_basepair(){
    // The returned basepair will always be the one already read and labeled "next"!
    // We will read the following basepair from dotplot
    basepair bp;
    if (!ready){
        bp.left = 0;
        bp.right = 0;
        bp.prob = 0.0;
    }else{
        bp = next;

        size_t2 left, right;
        std::string box;
        double prob;
	/*
	 * %This file contains the square roots of the base pair probabilities in the form
	 * % i  j  sqrt(p(i,j)) ubox
	 */

	if (fileHandle_ >> left >> right >> prob >> box)        // Try to read the next bp
            if (box == "ubox"){         // only ubox
                next.left = left;
                next.right = right;
                next.prob = prob * prob;    // Everything fine
            }else
                return read_basepair(); // recursively try again (it will also return "next") 
        else
            ready = false; // read FAILED = no basepairs left

    }
    return bp;
}


Dotplot_Wrapper::~Dotplot_Wrapper() {
}

