/* 
 * File:   Input_Reader.cc
 * Author: TCO
 * 
 * Created on 1. Oktober 2009, 12:39
 */

#include "Input_Reader.hh"

Input_Reader::Input_Reader(std::string fname) {
    file_.open(fname.c_str());
    seq_len_ = -1;
    max_span_ = -1; // NOTE TO ME. max_span is unsigned
    num_all_arcs_ = -1;
    num_seqs_ = 1;

    if (!file_.good()){
        std::cerr << "Cannot read "<< fname << std::endl;
	std::exit(-1);
    }

    if (!parse_header()){
        std::cerr << "Problems reading the header of input file " << fname << std::endl;
	std::exit(-1);
    }
}

bool Input_Reader::parse_header(){
    std::string line;
    while (!valid() && !file_.eof()){
        file_ >> line;

        if (line == "__SEQ_ID__")
            file_ >> seq_ID_;
        else if (line == "__NUM_SEQS__")
            file_ >> num_seqs_;
        else if (line == "__SEQ_LEN__")
            file_ >> seq_len_;
        else if (line == "__MAX_SPAN__")
            file_ >> max_span_;
        else if (line == "__NUM_ARCS__")
            file_ >> num_all_arcs_;
    }
    return !file_.eof();
}

/*
 * Checks whether the header has been read correct
 */
bool Input_Reader::valid(){
    return ( (!seq_ID_.empty()) && (seq_len_ > -1) && (max_span_ != -1) && (num_all_arcs_ > -1) );
}


size_t2 Input_Reader::seq_len(){ return seq_len_; }

size_t2 Input_Reader::num_all_arcs(){ return num_all_arcs_; }

size_t Input_Reader::num_seqs(){ return num_seqs_; }

std::string Input_Reader::seq_name(){ return std::string(seq_ID_); }

size_t Input_Reader::get_max_arc_span(){ return max_span_; }

Input_Reader::~Input_Reader() {
}

