/* 
 * File:   Input_Reader.hh
 * Author: TCO
 *
 * Created on 1. Oktober 2009, 12:39
 */

#ifndef _INPUT_READER_HH
#define	_INPUT_READER_HH

#include <fstream>
#include <iostream>
#include <string>
#include <vector>
#include <cstdio>
#include <cstdlib>
#include "base.hh"
#include "params.hh"

class Input_Reader {
public:
    Input_Reader(std::string fname);
    virtual ~Input_Reader();


    /*
     * returns vector of arcs with right end at i
     */
    virtual RightAdjList& right_adjlist(const size_t2 i) =0;

    /*
     * returns vector of arcs with left end at i
     */
    virtual LeftAdjList&  left_adjlist(const size_t2 i) =0;

    /*
     * return the i.th base of the sequence.
     * 1-based
     */
    virtual char operator [](const size_t2 i) =0;


    /*
     * return the query sequence identifier
    */
    virtual std::string seq_name();

    /*
     * returns the maximal span of ALL arcs (regardless of the minimal bp-probabiliy!)
    */
    virtual size_t get_max_arc_span();

    /*
    * returns the length of sequence
    */
    virtual size_t2 seq_len();

    /*
     * returns the count of all signifikant(!) arcs
     */
    virtual size_t2 num_arcs() =0;

    /*
     * returns the count of ALL available arcs
     */
    virtual size_t2 num_all_arcs();

    /*
     * returns the number of sequences in case of a multiple alignment
     */
    virtual size_t num_seqs();


protected:
    std::ifstream file_;
    std::string seq_ID_;
    size_t2 num_all_arcs_;
    size_t2 seq_len_;
    size_t max_span_;
    size_t num_seqs_;

    virtual bool parse_header();
    virtual bool valid();
};

#endif	/* _INPUT_READER_HH */

