/* 
 * File:   Input_Writer.cc
 * Author: TCO
 * 
 * Created on 29. September 2009, 11:16
 */

#include <fstream>
#include <iostream>
#include "Input_Writer.hh"
#include <algorithm>

bool bp_sort(const basepair a, const basepair b){
    return (a.right <= b.right) && (a.left <= b.left);
}

Input_Writer::Input_Writer(Input_Wrapper &wrapper):
    wrapper_(wrapper)
{
    max_span_ = 0;
    num_seqs_ = wrapper_.get_num_sequences();
    sequence_.resize(num_seqs_);
    seq_ID_   = wrapper_.get_seqID();
    for (int i=0; i<num_seqs_; i++)
        sequence_[i] = wrapper_.get_sequence();
    seq_len_  = sequence_[0].length();

    num_arcs_ = 0;
    fetch_basepairs();
}


/*
 * Fetch ALL basepairs one by one from the wrapper and sort them into bp_ list,
 * ordered by their RIGHT end
 * Determines max_span_ and num_arcs_
 */
void Input_Writer::fetch_basepairs(){
    bp_.resize(seq_len_+2);
    basepair bp;
    while (wrapper_.has_more_basepairs()){
        bp = wrapper_.read_basepair();
        if (bp.left >= bp.right)
            continue;
        bp_[bp.right].push_back(bp);
        max_span_ = max(max_span_, bp.right - bp.left);
        num_arcs_++;
    }


}

Input_Writer::~Input_Writer() {
    bp_.clear();
}

/*
 * Write the output file to fileHandle
 *
 */
bool Input_Writer::write_to(std::ofstream& fileHandle){
    if (!fileHandle.good())
        return false;

    fileHandle << "__SEQ_ID__ "   << seq_ID_   << std::endl;
    fileHandle << "__NUM_SEQS__ " << num_seqs_ << std::endl;
    fileHandle << "__SEQ_LEN__ "  << seq_len_  << std::endl;
    fileHandle << "__MAX_SPAN__ " << max_span_ << std::endl;
    fileHandle << "__NUM_ARCS__ " << num_arcs_ << std::endl;

    for (int i=1; i<=seq_len_; i++){
        
        for (int j=0; j<num_seqs_; j++)
            fileHandle << sequence_[j][i-1];

        fileHandle << " " << bp_[i].size() << " ";

        if (bp_[i].size() >= 2)
            std::sort(bp_[i].begin(), bp_[i].end(), bp_sort);

        for (int j=0; j < bp_[i].size(); j++)
            fileHandle << bp_[i][j].left << " " << bp_[i][j].prob << " ";
        fileHandle << std::endl;
    }
    
    fileHandle.close();
    return true;
}
