/* 
 * File:   ImportDotplot.cc
 * Author: TCO
 *
 * Created on 29. September 2009, 19:55
 */

#include <stdlib.h>
#include "Input_Writer.hh"
#include "Dotplot_Wrapper.hh"
#include "RNAplfold.hh"
#include "Fake_Input.hh"
#include <iostream>
#include <fstream>
#include "options.hh"
#include "MultiDotplot_Wrapper.hh"

bool opt_help;
bool opt_version;
bool opt_verbose;
bool process_success;

bool HAVE_INPUT;
bool HAVE_OUTPUT;
bool HAVE_FORMAT;

std::string input;
std::string add_input;
std::string output;
int format;


option_def my_options[] = {
			{"help",'h',&opt_help,O_NO_ARG,0,O_NODEFAULT,"","This help"},
			{"version",'V',&opt_version,O_NO_ARG,0,O_NODEFAULT,"","Version info"},
			{"verbose",'v',&opt_verbose,O_NO_ARG,0,O_NODEFAULT,"","Verbose"},

			{"",0,0,O_SECTION,0,O_NODEFAULT,"","Scoring parameters"},

			{"in",'i',0,O_ARG_STRING,&input,O_NODEFAULT,"Input file","Input file"},
			{"out",'o',0,O_ARG_STRING,&output,O_NODEFAULT,"Output file","YaSeTo Output file"},
			{"add",'a',0,O_ARG_STRING,&add_input,"","Additional Input","Addition Input file for input formats that require more than one inputfile"},
			{"format",'f',0,O_ARG_INT,&format,"1","input format","see FORMATS"},
                      

			{"",0,0,O_SECTION,0,O_NODEFAULT,"","FORMATS"},
                        {"1",0,&opt_help,O_NO_ARG,0,O_NODEFAULT,"","Simple Dotplot (e.g. -i query_dp.ps)"},
                        {"2",0,&opt_help,O_NO_ARG,0,O_NODEFAULT,"","Multiple Dotplot"},
                        {"3",0,&opt_help,O_NO_ARG,0,O_NODEFAULT,"","New RNAplfold output"},
                     
                        {"",0,0,0,0,O_NODEFAULT,"",""}
 };




/*
 * 
 */
int main(int argc, char** argv) {

    bool process_success=process_options(argc,argv,my_options);

    if (opt_help) {
	print_help(argv[0],my_options);
	exit(0);
    }

    if (opt_version) {
	cout << "1.0" <<endl<<endl;
	exit(0);
    }

    if (!process_success) {
	puts(O_error_msg);
	printf("USAGE: ");
	print_usage(argv[0],my_options);
	printf("\n");
	exit(-1);
    }

    if (opt_verbose)
	print_options(my_options);

    if (format > 1){
        if (add_input.empty()){
            printf("Format %i depends on more than one input files. Specify --add option\n", format);
            exit(-1);
        }
    }

    Input_Wrapper *wrapper;

    switch (format){
        case 3:
            wrapper = new RNAplfold(input, add_input);
            break;
        case 2:
            wrapper = new MultiDotplot_Wrapper(input, add_input);
            break;
        case 1:
            wrapper = new Dotplot_Wrapper(input);
            break;
        case 0:
            wrapper = new Fake_Input();
            break;


    }

    Input_Writer writer(*wrapper);
    std::ofstream file;
    file.open(output.c_str());
    writer.write_to(file);
    return (EXIT_SUCCESS); 



}

