/**********************************************************************
 *
 * Locarna-Scan: A modification of Locarna in terms of homology search
 *
 * Copyright (C) Michael Siebauer <michael@siebauer.com>
 *               2008-2011
 *
 * YaSeTo based on LocARNA: LOCal Alignment of RNA
 *
 * Copyright (C) Sebastian Will <will(@)informatik.uni-freiburg.de>
 *               2005, 2006
 *
 ***********************************************************************/

// TODO: Konstanten und Kompileoptionen aus config.h lesen + configure Skript schreiben


#include <iostream>
#include <algorithm>
#include <vector>
#include <string>
#include "time.h"
#include "params.hh"
#include "ribosum.hh"
#include "aligner.hh"
#include "ScoringModell.hh"
#include "options.hh"
#include "Query.hh"
#include "evalueCALC.h"
#include <queue>
#include "Target.hh"
#include "base.hh"

#include <fstream>



using namespace std;

const string VERSION_STRING="Locarna-Scan 1.9";

double min_prob;          // only pairs with a probability of at least min_prob are taken into account
double query_prob;        // minimal arc prob. for query, by default same as min_prob
score_t match_score;      // score for matching base pair
score_t mismatch_score;   // score for mismatching base pait
score_t gap_open_score;   // score for opening a new gap
score_t gap_extend_score; // score for extending gap

//TODO: Replace by e-value cutoff
double  e_cutoff = -1.0;  // E-value cutoff
int     max_span = -1;    // Maximal allowed arc span
score_t structure_weight; // score for an arc
score_t tau_factor;       // contribution of sequence similarity in an arc match (in percent)
score_t max_diff;         // maximal difference for positions of alignment edges
                          // (only used for ends of arcs)
score_t max_diff_am = -1; //maximal difference between two arc ends, -1 is off
score_t min_pos, max_pos;
string region;

time_t start, finish;

bool opt_help;
bool opt_version;
bool opt_verbose;
bool show_alis; // Show the full alignments     ! Huge Slow Down!
bool sequ_local; // Do a query-local alignment

string ribosum_file;       // Ribosum matrix
string output_file;        // Output file for ALL scores
string training_file;      // Output the score distribution
bool use_ribosum;
bool opt_logoddscores;

// ------------------------------------------------------------
// File arguments
string bpsfile1 = ""; // Query
string bpsfile2 = ""; // Target

// ------------------------------------------------------------
//
// Options
//



option_def my_options[] = {
			{"help",'h',&opt_help,O_NO_ARG,0,O_NODEFAULT,"","This help"},
			{"version",'V',&opt_version,O_NO_ARG,0,O_NODEFAULT,"","Version info"},
			{"verbose",'v',&opt_verbose,O_NO_ARG,0,O_NODEFAULT,"","Verbose"},

			{"",0,0,O_SECTION,0,O_NODEFAULT,"","Scoring parameters"},

			{"match",'m',0,O_ARG_INT,&match_score,"50","score","Match score"},
			{"mismatch",'M',0,O_ARG_INT,&mismatch_score,"0","score","Mismatch score"},
			{"ribosum-file",0,0,O_ARG_STRING,&ribosum_file,"Matrices/RIBOSUM85_60","f","Ribosum file"},
			{"use-ribosum",0,0,O_ARG_BOOL,&use_ribosum,"false","bool","Use ribosum scores"},
			{"logoddscores",0,0,O_ARG_BOOL,&opt_logoddscores,"false","bool","Score by log odd scores computed from query alignment."},
			{"indel",'i',0,O_ARG_INT,&gap_extend_score,"-350","score","Indel score"},
			{"indel-opening",0,0,O_ARG_INT,&gap_open_score,"-500","score","Indel opening score"},
			{"struct-weight",'s',0,O_ARG_INT,&structure_weight,"200","score","Maximal weight of 1/2 arc match"},
			{"tau",'t',0,O_ARG_INT,&tau_factor,"0","factor","Tau factor in percent"},

			{"",0,0,O_SECTION,0,O_NODEFAULT,"","Heuristics for speed accuracy trade off"},

			{"min-prob",'p',0,O_ARG_DOUBLE,&min_prob,"0.0005","prob","Minimal Target arc probability"},
                        {"query-arc-prob",'q',0,O_ARG_DOUBLE,&query_prob,"-1","prob","Minimal Query arc probability (if not explicity set equals Target arc prob (-p))"},
			{"max-diff-am",'D',0,O_ARG_INT,&max_diff_am,"-1","diff","Maximal difference for sizes of matched arcs (default: off)"},
                        {"span",'S',0,O_ARG_INT,&max_span,"-1","diff","Maximal arc span (default: off)"},
                        {"e-cutoff",'e',0,O_ARG_DOUBLE,&e_cutoff,"-1.0","cutoff","Show only hits below this e-value cutoff (default: off)"},

			{"",0,0,O_SECTION,0,O_NODEFAULT,"","Special sauce options"},
			{"output-file",0,0,O_ARG_STRING,&output_file,"","f","Output all alignment scores to this file"},
                        {"training-file",0,0,O_ARG_STRING,&training_file,"","f","Output the score distribution used for fitting to this file"},
                        {"query-local",0,0,O_ARG_BOOL,&sequ_local,"false","bool","Do a Query local alignment (like locarna)"},
			{"region",'R',0,O_ARG_STRING,&region,"","begin:end","Do an alignment only for the subregion <begin:end>"},
                        {"show-alis",0,0,O_ARG_BOOL,&show_alis,"false","bool","Display the full alignments (real slow down!)"},

            		{"",0,0,O_SECTION,0,O_NODEFAULT,"","RNA sequences and pair probabilities"},

                        {"",0,0,O_ARG_STRING,&bpsfile1,O_NODEFAULT,"query","QUERY input file"},
			{"",0,0,O_ARG_STRING,&bpsfile2,O_NODEFAULT,"targets","list of TARGET files"},
			{"",0,0,0,0,O_NODEFAULT,"",""}
 };




// ------------------------------------------------------------
// MAIN


// TODO: EIGENE DATENSTRUKTUREN FÜR DIE QUERY SEQUENZ BASTELN?!

int main(int argc, char **argv) {
    time(&start);
    bool process_success=process_options(argc,argv,my_options);

    if (opt_help) {
	cout << VERSION_STRING<<endl<<endl;
	print_help(argv[0],my_options);
	exit(0);
    }

    if (opt_version) {
	cout << VERSION_STRING<<endl<<endl;
	exit(0);
    }

    if (!process_success) {  //TODO: Wenn jemand weiss wie ich die Target parameter richtig parsen kann dann bitte.
        if (strcmp(O_error_msg, "Too many arguments.\n") != 0){
                puts(O_error_msg);
                printf("USAGE: ");
        	print_usage(argv[0],my_options);
                printf("\n");
                exit(0);
        }
    }

    if (opt_verbose)
	print_options(my_options);

    if (region.empty() == false){
	size_t colon = region.find(':');
	if (colon != std::string::npos){
		min_pos = atoi(region.substr(0, colon).c_str());
		max_pos = atoi(region.substr(colon+1).c_str());
		}else{
		print_help(argv[0],my_options);
	        exit(0);
	}
    }


    // Delete old output file if present
    std::remove(output_file.c_str());

    cout << "#";
    for (int i = 0; i < argc; i++)
	cout << argv[i] << " ";
    cout << std::endl;


    if (query_prob == -1)               // If unset, take the Target arc prob cutoff instead
        query_prob = min_prob;

    BasePairsParams bpp_query(structure_weight, query_prob);
    BasePairsParams bpp_target(structure_weight, min_prob);
    Query query(bpsfile1, bpp_query);

    // Alles was länger als 2 x Query ist, macht keinen Gewinn mehr!
    if (max_span == -1)
        max_span = 2 * query.get_max_arc_span();

    // Wenn eine Region angegeben ist, duerfen arcs nicht darueber hinaus
    if (!region.empty())
        max_span = min(max_span, (max_pos - min_pos));

    cout << "#Query Name:\t"  << query.seq_name()  << std::endl;
    cout << "#Query Length:\t"  << query.seq_len()  << std::endl;

    if (query.num_seqs() > 1)
        cout << "#Query is multiple alignment with " << query.num_seqs() << " sequences" << std::endl;
    cout << "#Query Arcs:\t" << query.num_arcs() << " of " << query.num_all_arcs() <<endl;


    AlignerParams aligner_params(match_score, mismatch_score, gap_open_score, gap_extend_score,
				 tau_factor, max_diff, max_diff_am, max_span);


    if (opt_logoddscores && use_ribosum) {
	use_ribosum=false;
	std::cerr << "WARNING: option logoddscores overrides conflicting option use-ribosum" << std::endl; 
    }
    
    // Do we have a ribosum matrix?
    RibosumFreq * ribo = (use_ribosum)? new RibosumFreq(ribosum_file) : NULL;

    vector<unsigned int> pos_counts;
    vector<unsigned int> neg_counts;
    PriorityQueue *k_best_list = new PriorityQueue;

    vector<string> target_list; // Split list into single Target files
    target_list.push_back(bpsfile2); // I know, its a stupid workaround
    for (int index = optind; index < argc; index++) // copy argv target arguments into vector
        target_list.push_back(argv[index]);
    cout << "#Got " << target_list.size() << " targets" << endl;
    vector<string> target_names;
    for (size_t target_number = 0; target_number < target_list.size(); target_number++){
        Target target(target_list[target_number], bpp_target); // Load Target Sequence
        target_names.push_back(target.seq_name());              // Store the Sequence Name for later output
        target.set_max_span(min(target.get_max_arc_span(), (size_t)max_span)); // Der max span kann nochmals reduziert sein, wenn es gar keine so grossen arcs gab!
        cout << "#Target: " << target.seq_name() << " length: " << target.seq_len() << " total arcs: " << target.num_all_arcs() << " max. arc span: " << target.get_max_arc_span() << endl;

        // Load Scoring Model
        ScoringModell scorer(query, target, aligner_params, ribo, opt_logoddscores);
	
        // Load Aligner Model
        Aligner aligner(scorer, query, target, output_file, show_alis);
        if (sequ_local)         // local alignment?
            aligner.enable_query_local_alignment();

        // >>> START ALIGNER <<<
        score_t score = (region.empty() ? aligner.local_align() : aligner.local_align(min_pos, max_pos) );
        cout << "#Target Arcs > min. arc prob.:\t" << target.num_arcs() << endl;
        PriorityQueue & tmp_kbest = aligner.get_k_best();     // get K-Besten Alignment
        // Fuege die neuen K-besten der globalen liste hinzu
        while (!tmp_kbest.empty()){
            result_type tmp = tmp_kbest.top();
            tmp.target_number = target_number;          // + information von welchem Target sie stammten
            k_best_list->push(tmp);
            tmp_kbest.pop();
            score = tmp.score;

            if (score < 0){                             // Zahle nebenbei die haeufigkeit der scores
                score = -score;
                if (score >= (score_t)neg_counts.size())
                    neg_counts.resize(score+1, 0);
                neg_counts[score]++;
            }else{
                if (score >= (score_t)pos_counts.size())
                    pos_counts.resize(score+1, 0);
                pos_counts[score]++;
            }
        }
    }

    // Die Aufspaltung der Scores in 2 Listen war noetig, weil ich den minimalen Score nicht vorher kenne
    int count_max = 0;
    int x_max = 0;
    // Ich suche jetzt den haeufigsten Score -> x_max

    for (size_t i=0; i < pos_counts.size(); i++)
        if ((int)pos_counts[i] > count_max){
            count_max = pos_counts[i];
            x_max = i;
        }
    for (size_t i=0; i < neg_counts.size(); i++)
        if ((int)neg_counts[i] > count_max){
            count_max = neg_counts[i];
            x_max = -i;         // -i weil score ja negativ ist.
        }

    
    score_t min_score = (neg_counts.size() == 0)?0:-(neg_counts.size()-1);
    score_t max_score = (pos_counts.size() == 0)?0:pos_counts.size()-1;
    
    // TODO:
    // the following modification of x_max is ad hoc and violates the semantics of x_max
    // as "most observed score"
    x_max = max_score - 150*abs(x_max - max_score)/100;
    if (x_max<min_score) x_max=min_score;
    
    unsigned int num_scores = abs(x_max - max_score);
    cout << "#Score Distribution:" << endl;
    cout << "#min_score: " << min_score << " max_score: " << max_score << endl;
    cout << "#Most observed score: " << x_max << " remaining scores after filtering: " << num_scores << endl;

    // Jetzt werden nur noch alle Scores >= xmax aufgehoben und die ScoreVerteilung verschoben! x_max --> 0!
    unsigned int counts[num_scores];
    for (size_t i=0; i < pos_counts.size(); i++){
        if ((int)i < x_max)  continue;
        counts[i-x_max] = pos_counts[i];
    }
    for (int i=0; i < (int)neg_counts.size(); i++){
        if (-i < x_max) continue;
        counts[-i-x_max] = neg_counts[i];
    }
    pos_counts.clear();
    neg_counts.clear();

    // Sliding Average funktioniert noch besser!
    size_t BIN_SIZE = 100;    // passt ganz gut, hat aber keinerlei math. rechtfertigung
    unsigned int bins[num_scores];
    long average = 0;
    queue<unsigned int> score_window;
    for (size_t i = 0; i < num_scores; i++){
        average += counts[i];
        score_window.push(counts[i]);
        if (score_window.size() == BIN_SIZE){
            if (average > (int)BIN_SIZE)                     // kommt sonst zu boesen rundungsfehlern
                bins[i-BIN_SIZE] = average / BIN_SIZE;  // float div besser?
            else
                bins[i-BIN_SIZE] = 0;
            average -= score_window.front();
            score_window.pop();
        }
    }
    // Restlichen Scores auch noch ausgeben oder hoer ich hier einfach auf?


    /* SCORE FITTING UND EWERT BERECHNUNG */
    // For Janas E-Value calculation convert to Log Score and split into 2 arrays
    float *logdata[2];
    logdata[0] = new float[num_scores]; // x and y values
    logdata[1] = new float[num_scores]; // x and y values
    int k=0;    // pos_index in logdata
    for (size_t i=0; i<num_scores; i++){   // split into x and y and remove 0-entries
        if (bins[i] == 0) continue;
        logdata[0][k] = i;
        logdata[1][k] = log(bins[i]);
        k++;
    }

    // Output training data if wanted -> These are the binned scores
    if (!training_file.empty()){
        string bin_training(training_file);
        bin_training.append("_BINS");
        ofstream out(bin_training.c_str());
        for (int i=0; i < k; i++){
            out << logdata[0][i] << "\t" << logdata[1][i] << endl;
        }
        out.close();
    }

    // Get the fitting Range
    int left_score,  right_score;
    bool no_evalues = false;
    
    if(true) {
	cout << "#E-value estimation disabled!" << endl;
        no_evalues = true;
    } else if (k < 100){        // arbitrary small value !
        cout << "#Sorry, too less scores to do an usefull e-value estimation!" << endl;
        no_evalues = true;
    }else{
        cout << "#Do Score Fitting" << endl;
        gotoh_getFitRange(logdata, k, &left_score, &right_score);
        cout << "#Left score range: " << left_score << endl;
        cout << "#Right score range: " << right_score << endl;
    }



     // Fuer die E-Wert Berechnung nehm ich aber die echten Scores nicht die aus den bins?
    k=0;    // pos_index in logdata
    for (size_t i=0; i<num_scores; i++){   // split into x and y and remove 0-entries
        if (counts[i] == 0) continue;
        logdata[0][k] = i;
        logdata[1][k] = log(counts[i]);
        k++;
    }

    // Output training data if wanted -> These are the real (shiftet) scores
    if (!training_file.empty()){
        ofstream out(training_file.c_str());
        for (int i=0; i < k; i++){
            out << logdata[0][i] << "\t" << logdata[1][i] << endl;
        }
        out.close();
    }


    // Schmeiss die scores raus die kleiner als der am haeufigsten gesehene sind
    PriorityQueue neu;        // Das wird die neue globale Liste ohne die zu kleinen scores!
    while (!k_best_list->empty()){
        if (k_best_list->top().score >= x_max)
            neu.push(k_best_list->top());
        k_best_list->pop();
    }
    k_best_list = &neu;

    double evalues[k_best_list->size()];                        // Hier kommen die E-Werte rein, Reihenfolge entspricht der von k_best_list
    score_t k_best_tmp[k_best_list->size()];                    // Hier kommen nur die (verschobenen) Scores rein fuer die jetzt e-Werte berechnet werden
    PriorityQueue copy(*k_best_list);                           // eine kopie der globalen kBestenListe weil ich aus ihr nicht lesen kann ohne sie zu leeren.
    int max_evalue_index = 0;                                   // Output only scores right of die index (0 == all scores)

    for (size_t i=0; i<k_best_list->size(); i++){ // Copy the kbest scores into own array + shift
        k_best_tmp[i] = copy.top().score - x_max;
        //k_best_tmp[i] = copy.top().score -min_score;
        copy.pop();
    }


    if (!no_evalues){
        gotoh_calcEvalue(evalues, k_best_tmp, k_best_list->size(), logdata, left_score, right_score);

        // Remove all kbest which are left of the maximal evalue
        // Problem, e-values can be NAN
        for (size_t i=0; i<k_best_list->size(); i++)
            if (evalues[i] > evalues[max_evalue_index])
                max_evalue_index = i;

        // Check whether fitting Process failed!
        for (size_t i=max_evalue_index+1; i<k_best_list->size(); i++)
            if (evalues[i-1] < evalues[i]){ // Evalue of higher score is worse --> Fitting failed!
                cout << "#E-Value fitting failed for some scores. I'll correct this manually..." << endl;
                for (size_t ii = i; ii < k_best_list->size(); ii++){
                    cout << "#Corrected " << evalues[ii] << " to ";
                    evalues[ii] = evalues[i-1] - (evalues[ii] - evalues[i-1]);  // TODO: Assure that evalue does not get negative!!
                    cout << evalues[ii] << endl;
                }
                break;
            }

        cout << endl;
        if (e_cutoff > -1)
            cout << "#Setting evalue cutoff to: " << e_cutoff << endl;
        else
            cout << "#Setting evalue cutoff to: " << evalues[max_evalue_index] << endl;
    }

    cout << "#Score\ttarget\tstart\tEnd\te-value" << endl;
    int i = 0;
    while (!k_best_list->empty()){
        result_type tmp = k_best_list->top();
        k_best_list->pop();
        if (i >= max_evalue_index){ // Show only kbest right of the maximal evalue!
                //if ((e_cutoff > -1) && (evalues[i] > e_cutoff))
                  //  continue;
            if (no_evalues)
                cout << tmp.score << "\t" << target_names[tmp.target_number] << "\t" << tmp.start << "\t" <<  tmp.stop << "\t" << "N.A."<< endl;
            else
                cout << tmp.score << "\t" << target_names[tmp.target_number] << "\t" << tmp.start << "\t" <<  tmp.stop << "\t" << evalues[i] << endl; // e-value
            if (show_alis)
                pretty_print_komprimiert(tmp.alignment, tmp.start, tmp.stop);
        }
        i++;
    }

    time(&finish);
    double diff = difftime(finish, start);
    int h = diff / 3600;
    int m = (diff - 3600*h) / 60;
    int s = diff - (h * 3600 + m * 60);
    cout << "#Elapsed time: " << h << "h " << m << "m " << s << "s" << endl;
    exit(0);
}



