/* 
 * File:   MultiDotplot_Wrapper.cc
 * Author: TCO
 * 
 * Created on 11. Oktober 2009, 19:16
 */


#include "MultiDotplot_Wrapper.hh"

MultiDotplot_Wrapper::MultiDotplot_Wrapper(std::string fName1, std::string fName2) {
    //seq_id_ = seqname_from_filename(fname);
    fasta_   = NULL;
    dotplot_ = NULL;
    i_ = 0;

    if (!identify_file_types(fName1.c_str(), fName2.c_str())){
        std::cerr << "Cannot identify or read input files" << std::endl;
	std::exit(-1);
    }else{
        fetch_sequence();
        ready = prepare_reading(); // search for the basepairs location
    }
}

std::string MultiDotplot_Wrapper::get_seqID(){
    return std::string("Multiple-Alignment");
}

std::string MultiDotplot_Wrapper::get_sequence(){
    return sequences_[i_++ % sequences_.size()];
}

bool MultiDotplot_Wrapper::identify_file_types(const char *f1, const char *f2){
    std::ifstream test(f1);
    std::string line;
    if (!test.good())
        return false;
    test >> line;
    if (line == "%!PS-Adobe-3.0")
        dotplot_ = new std::ifstream(f1);
    else if (line == "CLUSTAL")
        fasta_ = new std::ifstream(f1);

    std::ifstream test2(f2);
    if (!test2.good())
        return false;
    test2 >> line;
    if (line == "%!PS-Adobe-3.0")
        dotplot_ = new std::ifstream(f2);
    else if (line == "CLUSTAL")
        fasta_ = new std::ifstream(f2);

    return  ( (dotplot_ != NULL) && (fasta_ != NULL) );
}

bool MultiDotplot_Wrapper::has_more_basepairs(){
    return ready;
}



basepair MultiDotplot_Wrapper::read_basepair(){
    // The returned basepair will always be the one already read and labeled "next"!
    // We will read the following basepair from dotplot
    basepair bp;
    if (!ready){
        bp.left = 0;
        bp.right = 0;
        bp.prob = 0.0;
    }else{
        bp = next;

        size_t2 left, right;
        std::string box, hsb;
        double prob, junk;
	/*
	 * %This file contains the square roots of the base pair probabilities in the form
	 * % ??? ??? hsb i  j  sqrt(p(i,j)) ubox
	 */

	if (*dotplot_ >> junk >> junk >> hsb >>  left >> right >> prob >> box)        // Try to read the next bp
            if (box == "ubox"){         // only ubox
                next.left = left;
                next.right = right;
                next.prob = prob * prob;    // Everything fine
            }else
                return read_basepair(); // recursively try again (it will also return "next")
        else
            ready = false; // read FAILED = no basepairs left

    }
    return bp;
}



void MultiDotplot_Wrapper::fetch_sequence(){
        fasta_->seekg(std::ios_base::beg); // Just in case assure that we read the file from beginning
        std::string tmp, tmp_id, tmp_seq;
        std::map<std::string, std::string> tmp_map;
        while (!fasta_->eof()){
            std::getline(*fasta_, tmp);
            if (tmp.find("CLUSTAL") != std::string::npos)  // Check for headerline -> discard
                continue;
            if (tmp.find("*") != std::string::npos)  // Check for consensus line -> discard
                continue;
	    if (tmp.find_first_not_of(" \t") == std::string::npos) // check if line is completely empty
		continue;
            int end_id = tmp.find_first_of(" \t");
            if (end_id == std::string::npos)
                continue;
            tmp_id = tmp.substr(0, end_id);
            int start_seq = tmp.find_first_not_of(" \t", end_id);
            int end_seq = tmp.find_first_of(" \t\n", start_seq);
            tmp_seq= tmp.substr(start_seq, end_seq-start_seq);
            tmp_map[tmp_id].append(tmp_seq);
        }

        for (std::map<std::string, std::string>::iterator it = tmp_map.begin(); it != tmp_map.end(); it++)
            sequences_.push_back(it->second);        
}

size_t MultiDotplot_Wrapper::get_num_sequences(){
    return sequences_.size();
}

/*
 * Search for the basepair position in the dotplot and return the success
 * Also read ne NEXT basepair.
 *
 */
bool MultiDotplot_Wrapper::prepare_reading(){
    dotplot_->seekg(std::ios_base::beg);
    std::string buf;
    std::ios_base::streampos pos;
    while (!dotplot_->eof()){
        pos = dotplot_->tellg();
        std::getline(*dotplot_, buf);
        if ( (buf.find("hsb") != std::string::npos) && (buf.find("ubox") != std::string::npos) )
                break;        
    }

    if (dotplot_->eof()){
        std::cerr << "Problems reading the dotplot!" << std::endl;
        return false;
    }
    else{
        dotplot_->seekg(pos);
        ready = true;
        read_basepair();    // read the next basepair. The current value will be discarded. Should be uninitialized anyway
    }
    return ready;
}



MultiDotplot_Wrapper::~MultiDotplot_Wrapper() {
}





