/*
 * File:   PSSM.cc
 * Author: TCO
 *
 * Created on 20. April 2010, 09:59
 */

#include "PSSM.hh"

PSSM::PSSM(Query &query_) {
    bp_A = new double[query_.seq_len() + 1];
    bp_C = new double[query_.seq_len() + 1];
    bp_G = new double[query_.seq_len() + 1];
    bp_U = new double[query_.seq_len() + 1];
    bp_GAP = new double[query_.seq_len() + 1];

    const int N = query_.num_seqs();

    //double pseudocount = 0.0; // to be implemented --> problematic if column is completely gapped
  double pseudocount = std::sqrt(N); // GIBBS Style //
    //double pseudocount = 0.1; // MEME Style

    // basecounts
    double all_A = 0.0;
    double all_C = 0.0;
    double all_G = 0.0;
    double all_U = 0.0;

    // TODO: GAP KOSTEN BEDENKEN!!!

    // background probabilities
    double back_A = 0.25;
    double back_C = 0.25;
    double back_G = 0.25;
    double back_U = 0.25;

    basecount *bc;
    for (size_t2 i = 1; i <= query_.seq_len(); i++){
        bc = query_.get_basecount(i);
        all_A += bc->A;
        all_C += bc->C;
        all_G += bc->G;
        all_U += bc->U;
    }

    // Hintergrund verteilung berechnen
    const double all = all_A + all_C + all_G + all_U;
    
    assert(all>0);

    back_A = all_A / all;
    back_C = all_C / all;
    back_G = all_G / all;
    back_U = all_U / all;

    // baseprobs berechnen
    for (size_t2 i = 1; i <= query_.seq_len(); i++){  // TODO: Sollen die Gaps nun mitzaehlen oder nicht?!
        bc = query_.get_basecount(i);           // Muss man das noch Logarithmieren?!
        bp_A[i] = double(bc->A + pseudocount * back_A) / (N + pseudocount - bc->GAP);
        bp_C[i] = double(bc->C + pseudocount * back_C) / (N + pseudocount - bc->GAP);
        bp_G[i] = double(bc->G + pseudocount * back_G) / (N + pseudocount - bc->GAP);
        bp_U[i] = double(bc->U + pseudocount * back_U) / (N + pseudocount - bc->GAP);
	
        bp_GAP[i] = double(bc->GAP) / N;

        // normalize score?!
      /*  bp_A[i] /= back_A;
        bp_C[i] /= back_C;
        bp_G[i] /= back_G;
        bp_U[i] /= back_U;*/

        // Log-odd scores? --> Ist bei der RiboSum Berechnung drin
    }
    
    /*
    // DEBUGGING: print profile
    //
    std::cout << std::endl;
    std::cout << "----------------------------------------" << std::endl;
    for (size_t2 i = 1; i <= query_.seq_len(); i++){
	std::cout
	    << i << " "
	    << bp_A[i] << " "
	    << bp_C[i] << " " 
	    << bp_U[i] << " "
	    << bp_G[i] << " " 
	    << bp_GAP[i] << std::endl;
    }
    std::cout << "----------------------------------------" << std::endl;
    std::cout << std::endl;
    */
}


// TODO: replace char by BASE; POS unchecked!
double PSSM::base_prob(char c, size_t2 pos) const{
    switch (c){
        case 'A': return bp_A[pos];
        case 'C': return bp_C[pos];
        case 'G': return bp_G[pos];
        case '-': return bp_GAP[pos];
        case 'T': ;
        case 'U': return bp_U[pos];
        default:  return 0.0;
    }
}


PSSM::~PSSM() {
}

