/* 
 * File:   Query.cc
 * Author: TCO
 * 
 * Created on 30. September 2009, 20:37
 */

#include <iostream>
#include "Query.hh"

//TODO: Contraints

Query::Query(std::string fname, BasePairsParams &bpp) :
    Input_Reader(fname),
    params_(bpp),
    sequence_()
{

    sequence_.resize(seq_len_ + 1, num_seqs_);
    basecount_ = new basecount[seq_len_ + 1];
    
    left_.resize(seq_len_ + 1);
    right_.resize(seq_len_ + 1);
    num_arcs_ = 0;

    if (!parse()){
        std::cerr << "Problems parsing the query file" << std::endl;
	std::exit(-1);
    }
}


char Query::get_consensus(basecount &count){
    int most = std::max( std::max(count.A,count.C), std::max(count.G, count.U));

    if (most == count.A && most != std::max(std::max(count.C, count.G), count.U))
        return 'A';
    else if (most == count.C && most != std::max(std::max(count.A, count.G), count.U))
        return 'C';
    else if (most == count.G && most != std::max(std::max(count.C, count.A), count.U))
        return 'G';
    else if (most == count.U && most != std::max(std::max(count.C, count.G), count.A))
        return 'U';
    else
        return 'N';
}


bool Query::parse(){
    char base;
    int nArcs, right, left;
    float prob;
    score_t weight;

    right = 0;

    while (!file_.eof() && right < seq_len_){
        right++;
        basecount count = {0,0,0,0,0};
        for (int i=0; i<num_seqs_; i++){
            file_ >> base;
            sequence_(right, i) = base;
            switch (base){
	    case 'A': count.A++; break;
	    case 'C': count.C++; break;
	    case 'G': count.G++; break;
	    case 'U': count.U++; break;
	    case 'T': count.U++; break;
	    case '-': count.GAP++; break;
	    case 'N': count.A+=0.25; count.C+=0.25; count.G+=0.25; count.U+=0.25; break;
	    default: std::cerr << "Unknown/Unsupported base " << base << std::endl;
	    }
        }
        if (count.GAP == num_seqs_)
            std::cerr << "Bad alignment! Position " << right << " is completly gapped!" << std::endl;
            
        basecount_[right] = count;

        file_ >> nArcs;
        while (nArcs-- > 0){
            file_ >> left >> prob;
            if (prob >= params_.min_prob){
                weight = params_.probToWeight(prob, seq_len_);
                Arc arc = {num_arcs_++, left, right, weight};
                left_[left].push_back(arc);
                right_[right].push_back(arc);
            }
        }        
    }
   

    return ( (right == seq_len_) && (num_arcs_ <= num_all_arcs_) ); //  check if we really read all bases -> Could add more checks here
}

// 1-based!!!
char Query::operator[](const size_t2 i){
    assert(0 < i);
    assert(i <= seq_len_);
    return get_consensus(*get_basecount(i));
}

// 1-BASED!!!
char Query::base_at(const size_t2 pos, const int seq){
    assert(0 <= seq);
    assert(seq < num_seqs_);
    assert(0 < pos);
    assert(pos <= seq_len_ ); 
    return sequence_(pos, seq);
}


basecount * Query::get_basecount(size_t2 i){
    return &basecount_[i];
}


RightAdjList& Query::right_adjlist(const size_t2 i){
    return right_[i];
}

LeftAdjList& Query::left_adjlist(const size_t2 i){
    return left_[i];
}

size_t2 Query::num_arcs(){
    return num_arcs_;
}


Query::~Query() {
//    delete[] sequence_;
}
