/* 
 * File:   Query.hh
 * Author: TCO
 *
 * Created on 30. September 2009, 20:37
 */

#ifndef _QUERY_HH
#define	_QUERY_HH
#include "Input_Reader.hh"
#include "Input_Wrapper.hh"
#include "matrices.hh"

class Query : public Input_Reader{
public:
    Query(std::string fname, BasePairsParams &bpp);

    virtual ~Query();

    /*
     * returns the count of all signifikant(!) arcs
     */
    size_t2 num_arcs();

    bool is_multiple_alignment(void){
        return num_arcs_ > 1;
    }

    /*
     * returns vector of arcs with right end at i
     */
    RightAdjList& right_adjlist(const size_t2 i);

    /*
     * returns vector of arcs with left end at i
     */
    LeftAdjList&  left_adjlist(const size_t2 i);

    /*
     * return the i.th (CONSENSUS!!) base of the sequence.
     * 1-based
     */
    char operator [](const size_t2 i);
    
    /*
     * return the pos.th base of the seq.th sequence.
     * 1-based
     */
    char base_at(const size_t2 pos, const int seq);


    basecount *get_basecount(size_t2 i);



private:
    BasePairsParams &params_;
    Matrix<char> sequence_;
    basecount *basecount_;
    size_t2 num_arcs_;
    bool parse();
    char get_consensus(basecount &count);
    
    std::vector<RightAdjList> right_;
    std::vector<LeftAdjList> left_;
};

#endif	/* _QUERY_HH */

