/*
 * ScoringModell.cc
 *
 *  Created on: 18.07.2008
 *      Author: TCO
 */

#include "ScoringModell.hh"

ScoringModell::ScoringModell(Query &_query, Target &_target, AlignerParams &_ap, Ribosum *_ribo, bool _logoddscores) :
params(_ap),
query(_query),
target(_target),
ribo(_ribo),
logoddscores(_logoddscores),
pssm(_query)
{
    if (ribo != NULL) {
        char ca, cb;
        ribosum = static_cast<RibosumFreq *> (ribo);
        for (short a = 0; a < 5; a++) {
            ca = convert_i2b(a);
            for (short b = 0; b < 5; b++) {
                cb = convert_i2b(b);
                if (ribo->alphabet().in(ca) && ribo->alphabet().in(cb))
                    scores[a][b] = round2score(100.0 * ribosum->basematch_score_corrected(ca, cb));
                else
                    scores[a][b] = 0;
            }
        }

    } else if (!logoddscores) {
        for (int a = 0; a < 5; a++)
            for (int b = 0; b < 5; b++)
                if (a != 5 && b != 5)
                    scores[a][b] = (a == b) ? params.match_score : params.mismatch_score;
                else
                    scores[a][b] = 0;
    }

    // TODO: Constraints einlesen
    constraints = new bool[query.seq_len() + 1];
    for (int i=0; i<= query.seq_len(); i++)
        constraints[i] = false;
}

ScoringModell::~ScoringModell() {
}

//! return the base similarity

inline const score_t ScoringModell::sigma_logoddscores(size_t2 i, size_t2 j) const {
    float score = 100.0 * log(pssm.base_prob(target[j], i)*(float)4.0);
    // std::cout << i << " " << j << " "  <<pssm.base_prob(target[j], i) << " " <<score<<std::endl;
    return round2score(score);
}

inline const score_t ScoringModell::sigma_single(size_t2 i, size_t2 j) {
    return scores[convert_b2i(query[i])][convert_b2i(target[j])];
}

// TODO: Optimierung. Anstatt jedesmal zu pruefen, den Pointer-Link auf die Sigma-funktion direkt aendern!!
const score_t ScoringModell::sigma(size_t2 i, size_t2 j) {
    if (logoddscores) {
	return sigma_logoddscores(i,j);
    } else if (query.num_seqs() == 1) { 
        return sigma_single(i, j);
    } else {
        return sigma_multi(i, j);
    }
}

const score_t ScoringModell::gap_query(size_t2 i){
    return round2score(gap_extend_penalty() * (1.0 - pssm.base_prob('-', i)));
   // return gap_extend();
}

const score_t ScoringModell::gap_query_open(size_t2 i){
    //return round2score(gap_open_penalty()); 
    return round2score(gap_open_penalty() * (1.0 - pssm.base_prob('-', i))); // Gapöffnungen sollten auch preiswerter sein wenn dort Gaps wahrscheinlich sind!!
}




const score_t ScoringModell::sigma_test(size_t2 i, char tc) {
    const int t = convert_b2i(tc);

    return round2score(
            pssm.base_prob('A', i) * scores[convert_b2i('A')][t] +
            pssm.base_prob('C', i) * scores[convert_b2i('C')][t] +
            pssm.base_prob('G', i) * scores[convert_b2i('G')][t] +
            pssm.base_prob('U', i) * scores[convert_b2i('U')][t]);
}



const score_t ScoringModell::sigma_multi(size_t2 i, size_t2 j) {
    //int covered = 0; // todo: statisch deklarieren?!
    //double score = 0;
    const int t = convert_b2i(target[j]);

    return round2score(
            pssm.base_prob('A', i) * scores[convert_b2i('A')][t] +
            pssm.base_prob('C', i) * scores[convert_b2i('C')][t] +
            pssm.base_prob('G', i) * scores[convert_b2i('G')][t] +
            pssm.base_prob('U', i) * scores[convert_b2i('U')][t]);

/*    char* chars = query[i];
    for (int n = 0; n < query.num_seqs(); n++)
        if ((chars[n] != 'N') && (chars[n] != '-')) {
            covered++;
            score += scores[convert(chars[n])][t];
        }
*/
   // return round2score(score / covered); // TODO: lieber mit num_seqs arbeiten und covered ignorieren ?
}

score_t ScoringModell::arcmatch(const Arc & arcA, const Arc & arcB) {
    score_t sequence_contribution = 0;
    if (params.tau_factor != 0) {
        if (logoddscores) {
	    sequence_contribution = arcmatch_logoddscore(arcA, arcB);
	} else if (ribo != NULL) {
            sequence_contribution = ribosum_arcmatch_score(arcA, arcB);
        } else {
            sequence_contribution =
                    sigma(arcA.left, arcB.left) +
                    sigma(arcA.right, arcB.right);
        }
    }

    return
    // base match contribution
    // (also for arc-match add terms for the base match on both ends, weighted by tau_factor)
    round2score((params.tau_factor * sequence_contribution / 100)
            +
            // base pair weights
            (arcA.weight + arcB.weight));
}

score_t ScoringModell::ribosum_arcmatch_score(const Arc &arcA, const Arc &arcB) const {
    assert(ribo != NULL);
    double score = 0;

    char al, ar;
    const char bl = target[arcB.left];
    const char br = target[arcB.right];
    int considered_combinations = 0;

    const Alphabet<char> &alphabet = ribosum->alphabet();
    
    // compute average ribosum score
    for (size_t i = 0; i < query.num_seqs(); i++) { // run through all combinations of rows in A and B
        // wie sollen gaps behandelt werden?
        // current solution: ignore gap entries
       al = query.base_at(arcA.left, i);
       ar = query.base_at(arcA.right, i);
       if (al != '-' && ar != '-' && alphabet.in(al) && alphabet.in(ar)) {
	   considered_combinations++;

	   double score_contrib =
	       log(ribosum->arcmatch_prob(al, ar, bl, br)
		   /
		   (ribosum->basepair_prob(al, ar)
		     *
		     ribosum->basepair_prob(bl, br))
		   );

	   score += score_contrib;

       }
       
    }
    
    if (considered_combinations==0) return 0;

    score = 100.0 * score / considered_combinations;

    return round2score(score);
}



score_t ScoringModell::arcmatch_logoddscore(const Arc &arcA, const Arc &arcB) const {

    char al, ar;
    const char bl = target[arcB.left];
    const char br = target[arcB.right];
    
    float count = 0;
    size_t total = query.num_seqs();
    
    // count occurrence of bl br in input data
    for (size_t i = 0; i < query.num_seqs(); i++) { // run through all combinations of rows in A and B
       al = query.base_at(arcA.left, i);
       ar = query.base_at(arcA.right, i);
    
       // slightly too simple handling of Ns in query sequence
       if (al=='N' || ar=='N') {
	   count += 1/6.0;
       } else if (al == bl && ar == br) {
	   count ++;
       }
    }
    
    // add pseudocount and compute log odd
    
    assert(total>0);
    
    float pseudocount=0.1;//sqrt((float)total);
    
    float score = 200.0 * log( 
			      ((float)count+pseudocount/6.0)/(total + pseudocount) 
			      / (1.0/6.0) 
			       );
    
    // std::cout << arcA.left << " " << arcA.right << " " << bl << br << " " << count <<" " << total << " " <<score << std::endl;

    return round2score(score);
    
}

