/*
 * ScoringModell.hh
 *
 *  Created on: 18.07.2008
 *      Author: TCO
 */

#ifndef SCORINGMODELL_HH_
#define SCORINGMODELL_HH_
#include "params.hh"
#include "Query.hh"
#include "Target.hh"
#include <cassert>
#include "ribosum.hh"
#include <math.h>
#include "PSSM.hh"

class ScoringModell {
private:
    AlignerParams &params; //!< the parameter for the alignment
    Query &query;
    Target &target;
    bool * constraints;
    Ribosum *ribo;
    RibosumFreq *ribosum;
    bool logoddscores;

    score_t scores[5][5];
    PSSM pssm;
    const score_t sigma_logoddscores(size_t2 i, size_t2 j) const;
    const score_t sigma_single(size_t2 i, size_t2 j);
    const score_t sigma_multi(size_t2 i, size_t2 j);
public:

    const score_t sigma_test(size_t2 i, char t);


    // TODO: Auf enum umstellen

    short convert_b2i(const char c) const {
        switch (c) {
            case 'A':
                return 0;
            case 'C':
                return 1;
            case 'G':
                return 2;
            case 'T':
                return 3;
            case 'U':
                return 3;
            default:
                return 4;
        }
    }

    char convert_i2b(const short c) const {
        switch (c) {
            case 0:
                return 'A';
            case 1:
                return 'C';
            case 2:
                return 'G';
            case 3:
                return 'U';
            default:
                return 'N';
        }
    }


    ScoringModell(Query &query, Target &target, AlignerParams &ap, Ribosum *ribo, bool logoddscores);
    virtual ~ScoringModell();

    const score_t sigma(size_t2 i, size_t2 j);


    //! round a double to score_t.
    //! more robust than truncate

    inline score_t round2score(const double d) const {
        return (score_t) ((d < 0) ? (d - 0.5) : (d + 0.5));
    }


    //! contribution of sequence match/mismatch to the arc match score

    score_t tau(const size_t2 i, const size_t2 j, const size_t2 k, const size_t2 l) {
        return (params.tau_factor) ? (((sigma(i, j) + sigma(k, l)) * params.tau_factor) / 100) : 0;
    }

    score_t arcmatch(const Arc & arcA, const Arc & arcB);
    score_t ribosum_arcmatch_score(const Arc &arcA, const Arc &arcB) const;
    score_t arcmatch_logoddscore(const Arc &arcA, const Arc &arcB) const;


        // Kosten ein Gap an I zu erweitern!
        const score_t gap_query(size_t2 i);
        // Kosten ein Gap dort zu öffnen
        const score_t gap_query_open(size_t2 i);


        const score_t gap_query(const size_t2 from, const size_t2 to){
            return gap_query_open(from) + gap_extend_query(from, to);
        }

        const score_t gap_target(const size_t2 from, const size_t2 to){
            return gap_open_penalty() + gap_extend_target(from, to);
        }

	score_t gap_extend_query(const size_t2 from, const size_t2 to){
            score_t gap_sum = 0;
            for (size_t2 i = from; i <= to; i++){
                gap_sum += this->gap_query(i);
                if (constraints[i])
                    return INFTY;
            }
            return gap_sum;
	}

	inline score_t gap_extend_target(const size_t2 from, const size_t2 to){
		return params.gap_extend_score + (to - from) * params.gap_extend_score;
	}



	score_t max_diff_am(){
        return params.max_diff_am;
    }

    const inline score_t gap_open_penalty() const {
        return params.gap_open_score;
    }

    const inline score_t gap_extend_penalty() const {
        return params.gap_extend_score;
    }


};

#endif /* SCORINGMODELL_HH_ */
