/*
 * File:   Target.cc
 * Author: TCO
 *
 * Created on 1. Oktober 2009, 12:06
 */

#include "Target.hh"





Target::Target(std::string fname, BasePairsParams &bpp) :
    Input_Reader(fname),
    params_(bpp)
{    
    sequence_ = NULL;
}

void Target::re_init(void){
    win_size_ = max_span_ +1;    
    pos = 0;
    num_arcs_ = 0;

    if (sequence_ != NULL)
        free(sequence_);
    sequence_ = new char[win_size_];
    left_.clear();
    right_.clear();    
    left_.resize(win_size_);    
    right_.resize(win_size_);    
    parse(win_size_);    
}

void Target::set_max_span(size_t max_span){    
    max_span_ = max_span;       // Man kann nicht sicher genug sein 
    re_init();        
}


inline size_t Target::addr(size_t2 i){
    return (i % win_size_);
}

bool Target::parse(size_t2 count){
    char base;
    int nArcs, left;
    float prob;
    score_t weight;

    // Read next of "count" lines. The Target base and the number of arcs for this position
    while (!file_.eof() && count-- > 0 && file_ >> base >> nArcs){
        pos++;   // POS == RIGHT
        sequence_[addr(pos)] = base;
        // alte arcs loeschen
        right_[addr(pos)].clear();
        left_[addr(pos)].clear();
        while (nArcs-- > 0){    // Now read the following "nArcs" arcs
            file_ >> left >> prob;  
            if ((prob >= params_.min_prob) && ((pos - left) <= max_span_)){
                weight = params_.probToWeight(prob, seq_len_);
                Arc arc = {num_arcs_++, left, pos, weight};
                left_[addr(left)].push_back(arc);
                right_[addr(pos)].push_back(arc);
            }
        }
    }
    return (!file_.eof());
}

/* NEVER EVER TRY TO ACCESS POSITIONS THAT ARE PRIOR TO THE ACTUAL READING FRAME*/
char Target::operator[](const size_t2 i){
    assert(0 < i);
    assert(i <= seq_len_);       
    assert(i >= pos-win_size_);
    if (i > pos)
        parse(i - pos);
    /*if (i < pos-win_size_){
        std::cerr << "LeseFehler []" << std::endl;
        std::cerr << i << " pos " << pos << " win " << win_size_ << std::endl;
        exit(-1);
    }*/
    
    return sequence_[addr(i)];
}


RightAdjList& Target::right_adjlist(const size_t2 i){
    assert(i >= pos-win_size_);
    if (i > pos)
        parse(i - pos);
    /*if (i < pos-win_size_){
        std::cerr << "LeseFehler RA" << std::endl;
        std::cerr << i << " pos " << pos << " win " << win_size_ << std::endl;
        exit(-1);
    }*/
    return right_[addr(i)];
}

LeftAdjList& Target::left_adjlist(const size_t2 i){
    assert(i >= pos-win_size_);
    if (i > pos)
        parse(i - pos);
 /*   if (i < pos-win_size_){
        std::cerr << "LeseFehler LA" << std::endl;
        std::cerr << i << " pos " << pos << " win " << win_size_ << std::endl;
        exit(-1);
    }*/
    return left_[addr(i)];
}


size_t2 Target::num_arcs(){
    return num_arcs_;
}


Target::~Target() {
    delete[] sequence_;
}

