/* 
 * File:   Target.hh
 * Author: TCO
 *
 * Created on 1. Oktober 2009, 12:06
 */

#ifndef _TARGET_HH
#define	_TARGET_HH

#include "Input_Reader.hh"
#include <algorithm>


class Target : public Input_Reader {
public:
    /*
     * fname    .. fileName of the target structure
     * bpp      .. BasePairing Parameters
     * max_span .. Span of maximal allowed arc (<= 2x query length)
     */
    Target(std::string fname, BasePairsParams &bpp);
    
    void set_max_span(size_t max_span);   
    

    /*
     * return vector of arcs with right end at i
     */
    RightAdjList& right_adjlist(const size_t2 i);

    /*
     * return vector of arcs with left end at i
     */
    LeftAdjList&  left_adjlist(const size_t2 i);

    virtual ~Target();
    
    char operator [](const size_t2 i);

    size_t2 num_arcs();


private:

    BasePairsParams &params_;
    char * sequence_;
    size_t2 num_arcs_;
    size_t win_size_;
    size_t2 pos;
    bool parse(size_t2 count);
    void re_init(void); // Call me after max_span changed! 
    size_t addr(size_t2 i);
    std::vector<RightAdjList> right_;
    std::vector<LeftAdjList> left_;
};

#endif	/* _TARGET_HH */


