
#include "aligner.hh"
#include <cstdio>
#include <cstring>

// Redefine faster maximum functions
//#define max1(a, b) (((a) > (b))?(a):(b))
#define max1(a, b) std::max(a, b)
#define max2(a, b, c) (max1(a, std::max(b, c)))

Aligner::Aligner(ScoringModell &_scoring, Query &_query, Target &_target, std::string output_file, bool _show_alis=false) :
growing_score(false),   // Merkt sich ob der Score gerade ansteigend ist (erst später interessant)
locarna(false),         // for locarna like query-local alignments
alignment_strings(NULL),// erstmal allokieren wir nix. Ausser der User will die Alignments sehen.
tmp_alignment_strings(NULL),// erstmal allokieren wir nix. Ausser der User will die Alignments sehen.
show_alis(_show_alis),  // Keep track of the full alignments (not only start and end points)
max_ali_size(5 * _query.seq_len()),
query_len(_query.seq_len()),

query(_query),          // Query Arcs
target(_target),        // Target Arcs

histogram_file(output_file.c_str(),  std::ios_base::out | std::ios_base::app), // Hier landen ALLE Scores, wenn gewünscht (--output-file)
k_best_list(),          // Liste der k-besten Alignments
local_max_list(),       // speichere (temporaer) alle lokalen maxima (sollten im Worst case <= (query-laenge x k-best) sein
scoring(_scoring),      // Scoring Model

D(),                    // D matrix

MAX_SPAN(_target.get_max_arc_span()), // Maximale Arc Spannweite
NO_PROGRESS_OUTPUT(false) // Disable terminal progress output
{
    // Initialisierung der Matrizen
    M.resize(query_len + 1, MAX_SPAN + 2);   // Score Matrix   (hat volle m Zeilen aber rotiert ueber MAX_SPAN Spalten)
    // Mentale Notiz: Micha, nicht an den Grössen der OffsetMatrix rumspielen, sonst haut die initialisierung nicht hin
    tmp.resize(query_len+1, MAX_SPAN + 2);   // Tmp matrix fuer globale alignments zwischen arcs  ("echte Matrix")  TODO: Kann sie nicht eigentlich viel kleiner sein?!
    //D.resize(query.num_arcs(), MAX_SPAN);    // Der Wert ist eigentlich egal --> Die D Matrix wird automatisch größer bei Bedarf.

    tmp_del_target = new matrix_entry[query_len+1];
    // Wie lange ist ein Alignment maximal?
    // 2 x Query Laenge im worst case
    // Fuer jede Position brauch ich 2 Zeichen
    // und ein /0 Zeichen am Ende

    if (show_alis){
        // und das ganze fuer jeden Matrix  Eintrag == query_len + 1 * max_span + 2
        alignment_strings     = new char[(query_len + 1) * (MAX_SPAN + 2) * max_ali_size];
        tmp_alignment_strings = new char[(query_len + 1) * (MAX_SPAN + 2) * max_ali_size];
        // Jetzt muss ich jedem Matrix Element eine eindeutige Speicher Adresse im char Array geben.
        long pos = 0;
        for (int q = 0; q <= query_len; q++)
            for (int t = 0; t<= MAX_SPAN+1; t++){
                M(q,t).alignment   = &alignment_strings[pos];  // new char[max_ali_size];
                tmp(q,t).alignment = &tmp_alignment_strings[pos]; // new char[max_ali_size];
                pos += max_ali_size;
            }
        tmp_del_query.alignment = new char[max_ali_size];
        tmp_MAX.alignment = new char[max_ali_size];
        for (int i = 0; i <= query_len; i++)
            tmp_del_target[i].alignment = new char[max_ali_size];
    }

    last_score.score = 0;
    last_score.start = 0;
    last_score.stop  = 0;

    // Rand initialisieren
    for (size_t2 i = 0; i <= query_len; i++)
        tmp(i, MAX_SPAN+1).score = scoring.gap_query(i+1, query_len);       // Rechter Rand der Tmp Matrix bekommt Gap Strafen (! Rueckwaerts alignment!
    for (size_t2 i = 0; i <= MAX_SPAN; i++)
        tmp(query_len, i).score = scoring.gap_target(i, MAX_SPAN);  // Unterer Rand der Tmp Matrix bekommt Gap Strafen (globales Alignment rueckwaerts)
    tmp(query_len, MAX_SPAN+1).score = 0;
}

// Debug option. Makes an local alignment like locarna
void Aligner::enable_query_local_alignment(){ locarna = true; }

Aligner::~Aligner(){
    if (alignment_strings != NULL){
        delete[] alignment_strings;
        delete[] tmp_alignment_strings;
        delete[] tmp_del_query.alignment;
        delete[] tmp_MAX.alignment;
        for (int i = 0; i <= query_len; i++)
                delete[] tmp_del_target[i].alignment;
    }
}


void Aligner::store_score(const size_t2 start, const size_t2 end, const int score, char *ali) {
    if (histogram_file.good()){  // Output unfiltered scores
        histogram_file << target.seq_name() << " " << start << " " << end << " " << score << "\t" << end-start;
        if (ali != NULL)
          histogram_file << " " << ali;
        histogram_file << std::endl;
    }

    // DAS WAR DER LOKALE MAXIMA FILTER
    if ( (last_score.score > score) && (growing_score)){ // Score faellt wieder
        local_max_list.push(last_score);
        growing_score = false;
    } else if (score > last_score.score)                 // Es geht wieder aufwaerts
        growing_score = true;

    last_score.score = score;
    last_score.start = start;
    last_score.stop  = end;
    if (ali != NULL)
        last_score.alignment = std::string(ali);
}


// Computate and return the list of the k-best alignments (local_maxima with distance filter)
PriorityQueue & Aligner::get_k_best(){
    if (k_best_list.empty())
        find_kbest();
    return k_best_list;
}

// Return the list of local maximal alignments
PriorityQueue & Aligner::get_lm_list(){
    return local_max_list;
}

/*
 * Alle (ausreichend viele) lokalen Maxima liegen in tmp_best. Nun wollen wir nur
 * die mit mindestens Query-Abstand zueinander.
 */
void Aligner::find_kbest(){
    // Ich kopiere jetzt immer abwechselnd zwischen zwei Queues hin- und her und entferne
    // alle Eintraege die zu nah am aktuell besten Hits (Lokalen maximum) sind
    result_type next_best, i;
    std::deque<result_type> *zeiger_lesen = new std::deque<result_type>;
    std::deque<result_type> *zeiger_schreiben = new std::deque<result_type>;
    std::deque<result_type> *zeiger_tmp;


    // Unguenstigerweise ist meine Priority queue genau falschherum angeordnet...
    PriorityQueue copy(local_max_list);
    while(!copy.empty()){
        zeiger_lesen->push_front(copy.top());
        copy.pop();
    }

    zeiger_schreiben->clear();
    while (!zeiger_lesen->empty()){
        next_best = zeiger_lesen->front();      // Hole den aktuell besten Hit
        k_best_list.push(next_best);            // Den Besten nehmen wir auf jeden fall
        while (!zeiger_lesen->empty()){ // Hier gehts los.
            i = zeiger_lesen->front();
            zeiger_lesen->pop_front();
            // remove the alignment if it overlaps with the next_best alignment
            if (( i.stop < next_best.start) || (i.start > next_best.stop) )
                zeiger_schreiben->push_back(i);         // heben wir auf fuer spaeter
        }
        // zeiger wieder umschalten auf die andere Queue
        zeiger_tmp = zeiger_lesen;
        zeiger_lesen = zeiger_schreiben;
        zeiger_schreiben = zeiger_tmp;
    }
}


void pretty_print_komprimiert(const std::string ali, const size_t2 start, const size_t2 end){
    std::string arcs, s_query, target;
    bool query = true;
    for (std::string::const_iterator it = ali.begin(); it < ali.end(); it++){
        unsigned char c = *it; // !!!! UNSIGNED!
        if (query){
            if (c>126){
                arcs   += ')';
                c-=82;
            }else if (c>86){
                arcs   += '(';
                c-=41;
            }else
                arcs += ' ';
           s_query += c;
        }else
            target += c;
        query = !query;
    }
    if ((start == -1) || (end == -1))
        std::cout << std::endl << arcs << std::endl << s_query << std::endl << target << std::endl << arcs << std::endl << std::endl;
    else
        std::cout << std::endl << "\t" << arcs << std::endl << "0\t" << s_query << " " << s_query.length() << std::endl << start << "\t" << target << " " << end << std::endl << "\t" << arcs << std::endl << std::endl;
}


void pretty_print(const std::string ali, const size_t2 start=-1, const size_t2 end=-1){
    std::string arcs, s_query, target;
    for (size_t i=0; i<ali.size(); i++){
    //for (std::string::const_iterator it = ali.begin(); it < ali.end(); it++){
        //char c = *it;
        char c = ali[i];

        if ( (c == '(') || (c == '[') ){
            arcs   += '(';
            s_query  += ali[++i];//*(++it);
            target += ali[++i];//*(++it);
        }else if ( (c == ')') || (c == ']') )
            arcs[arcs.size()-1] = ')';
        else{
            s_query  += c;
            target += ali[++i];//*(++it);
            arcs   += ' ';
        }
    }

    if ((start == -1) || (end == -1))
        std::cout << std::endl << arcs << std::endl << s_query << std::endl << target << std::endl << arcs << std::endl << std::endl;
    else
        std::cout << std::endl << "\t" << arcs << std::endl << "0\t" << s_query << " " << s_query.length() << std::endl << start << "\t" << target << " " << end << std::endl << "\t" << arcs << std::endl << std::endl;
}


//* Legt fest ob ich diese Arc Kombination berechnen soll oder nicht
bool Aligner::lohnt_sich(const Arc &q, const Arc &t){
    if (scoring.max_diff_am() == -1)    // Wenn kein max diff filter angegeben, rechne immer alles aus!
        return (t.right - t.left) < 2 * (q.right - q.left);    // Die Target Sequenz darf nicht mehr als 2x Query lang sein. Ansonsten habe ich mehr Gaps als SequenzMatches -> sollte(!) sich nicht lohnen
    else
        return std::abs( (q.right - q.left) - (t.right - t.left)) <= scoring.max_diff_am();    // difference in arc-spans between q and t arc must be below cutoff
}


// Do a (reverse) global alignment of query[i..j] and target[k..l]
void Aligner::global_align(const size_t2 q_min, const size_t2 q_max, const size_t2 t_min, const size_t2 t_max) {
    score_t tmp1;
    tmp_arcsQuery = NULL;
    tmp_arcsTarget = NULL;
    for (size_t2 q = 0; q <= query_len; q++)
        tmp_del_target[q].score = -INFTY;

    // Adjust the matrix coords so that q_max+1, t_max+1 is the lower right index
    tmp.set_max_coord(q_max+1, t_max + 1);
    if (show_alis){     // Rand initialisieren
        tmp(q_max+1, t_max+1).alignment[0] = '\0';
        for (size_t2 t = t_max; t >= t_min; t--)
                sprintf(tmp(q_max+1, t).alignment, "-%c%s", target[t], tmp(q_max+1, t + 1).alignment);
        for (size_t2 q = q_max; q >= q_min; q--)
                sprintf(tmp(q, t_max+1).alignment, "%c-%s", query[q], tmp(q+1, t_max + 1).alignment);
    }

    for (size_t2 t = t_max; t >= t_min; t--) {
        tmp_del_query.score = -INFTY;
        tmp_arcsTarget = &target.left_adjlist(t);               // Liste aller Arcs mit Linkem Ende bei y

        for (size_t2 q = q_max; q >= q_min; q--) {
            tmp_arcsQuery = &query.left_adjlist(q);                       // NEEDLEMAN-WUNSCH REVERSE

            //del_query = max1(tmp(q + 1, t) + scoring.gap_query_open(q), del_query) + scoring.gap_query(q);
            if (tmp(q + 1, t).score + scoring.gap_query_open(q) > tmp_del_query.score){
                tmp_del_query.score = tmp(q + 1, t).score + scoring.gap_query_open(q);
                if (show_alis)
                    strcpy(tmp_del_query.alignment, tmp(q + 1, t).alignment);
            }
            tmp_del_query.score += scoring.gap_query(q); // gap_ext;     // alten query gap verlaengern

            //tmp_del_target[q - q_min] = max1(tmp(q, t + 1).score + scoring.gap_open_penalty(), tmp_del_target[q - q_min]) + scoring.gap_extend_penalty();
            if (tmp(q, t + 1).score + scoring.gap_open_penalty() > tmp_del_target[q - q_min].score){   // dasselbe fuer target
                tmp_del_target[q - q_min].score = tmp(q, t + 1).score + scoring.gap_open_penalty();
                if (show_alis)
                    strcpy(tmp_del_target[q - q_min].alignment, tmp(q, t+1).alignment);
            }
            tmp_del_target[q - q_min].score += scoring.gap_extend_penalty();

            //  MAX = max2(tmp(q + 1, t + 1) + scoring.sigma(q, t), tmp_del_target[q - q_min], del_query);
            tmp_MAX.score = max2(tmp(q + 1, t + 1).score + scoring.sigma(q, t),        // (mis)match
                                tmp_del_target[q-q_min].score,                                // insert
                                tmp_del_query.score);                                   // deletion

             if (show_alis){    // TODO:        Speicher Allokierung auch raus nehmen
		char s_tmp[std::max(strlen(tmp_del_query.alignment), strlen(tmp_del_target[q - q_min].alignment)) + 3];  // 2 Zeichen kommen hinzu + \0
                sprintf(s_tmp, "%c-%s", query[q], tmp_del_query.alignment);
                strcpy(tmp_del_query.alignment, s_tmp);
                sprintf(s_tmp, "-%c%s", target[t], tmp_del_target[q - q_min].alignment);
                strcpy(tmp_del_target[q - q_min].alignment, s_tmp);

                if (tmp_MAX.score == tmp_del_target[q - q_min].score)              // deletion
                    strcpy(tmp_MAX.alignment, tmp_del_target[q - q_min].alignment);
                else if (tmp_MAX.score == tmp_del_query.score)             // insert
                    strcpy(tmp_MAX.alignment, tmp_del_query.alignment);
                else
                    sprintf(tmp_MAX.alignment, "%c%c%s", query[q], target[t], tmp(q+1, t+1).alignment);
             }

            // Alle benoetigten D Eintraege sind schon bekannt und werden direkt benutzt (stimmt nicht!)
            for (LeftAdjList::const_iterator arcQ = tmp_arcsQuery->begin(); arcQ != tmp_arcsQuery->end() && arcQ->right <= q_max; ++arcQ)
                for (LeftAdjList::const_iterator arcT = tmp_arcsTarget->begin(); arcT != tmp_arcsTarget->end() && arcT->right <= t_max; ++arcT)
                    if (lohnt_sich(*arcQ, *arcT)){    // WICHTIG NICHT LOESCHEN!!!!!
                        //MAX = max1(MAX, D(arcQ->idx, arcT->idx) + tmp(arcQ->right + 1, arcT->right + 1));

                        tmp1 = D(arcQ->idx, arcT->idx).score + tmp(arcQ->right + 1, arcT->right + 1).score;
                         if (tmp1 > tmp_MAX.score){
                                tmp_MAX.score = tmp1;
                                if (show_alis){
                                        //sprintf(MAX.alignment, "[%c%c%s%c%c]%s", query[arcQ->left], target[arcT->left], D(arcQ->idx, arcT->idx).alignment,
                                        //                                          query[arcQ->right], target[arcT->right], tmp(arcQ->right + 1, arcT->right + 1).alignment);

                                    sprintf(tmp_MAX.alignment, "%c%c%s%c%c%s", query[arcQ->left]+41, target[arcT->left], D(arcQ->idx, arcT->idx).alignment,
                                                                                 query[arcQ->right]+82, target[arcT->right], tmp(arcQ->right + 1, arcT->right + 1).alignment);
                                    if (strlen(tmp_MAX.alignment) >= max_ali_size){
                                        std::cout << "  GLOBAL ALIGN " << q_min << " " << q_max << " vs "  << t_min << " " << t_max << std::endl;
                                        std::cout << q_max - q_min << " vs. " << t_max - t_min << std::endl;
                                        std::cout << "G" << q_min << ".." << q << ".." << q_max << " " << t_min << ".." << t << ".." << t_max <<std::endl;
                                        std::cout << "2SSTOP!!!" << max_ali_size << std::endl;
                                        std::cout<< tmp_MAX.alignment << " len " << strlen(tmp_MAX.alignment) << std::endl;
                                        pretty_print_komprimiert(tmp_MAX.alignment);


                                        for (int q = q_min; q<= q_max; q++)
                                                std::cout << query[q];
                                        std::cout << std::endl;
                                        for (int t = t_min; t<= t_max; t++)
                                                std::cout << target[t];
                                        std::cout << std::endl;

                                        std::cout << "qt + D + qt + tmp" << std::endl;
                                        std::cout << query[arcQ->left] << target[arcT->left] << "||" <<  D(arcQ->idx, arcT->idx).alignment;
                                        std::cout << "||" << query[arcQ->right] <<  target[arcT->right] << " + " <<  tmp(arcQ->right + 1, arcT->right + 1).alignment  << std::endl;
                                        std::cout << "D=";pretty_print_komprimiert(D(arcQ->idx, arcT->idx).alignment);
                                        std::cout << "tmp=";pretty_print_komprimiert(tmp(arcQ->right + 1, arcT->right + 1).alignment );

                                        exit(0);
                                    }
                                }
                         }
                    }
            tmp(q, t).score = tmp_MAX.score;

            if (show_alis)
                strcpy(tmp(q, t).alignment, tmp_MAX.alignment);
        }
    }
}





// Do a local Alignment between Target and Query
// ich laufe über alle positionen also nxm ...
// an jeder Position gibt es maximal 1/p_min Arcs ...
// ich berechne genau ein globales Alignment ...
//
// O( m x n x ( (1/p_min) x (1/p_min) + globales_alignment) )
// O(n^2) x (O(n^2) + O(n^4)) => ~O(n^6)
// target_min und target_max, bestimmen optional nur einen Teilbereich zu alignieren
score_t Aligner::local_align(size_t2 target_min, size_t2 target_max) {
    const size_t2 target_len = (target_max == -1) ? target.seq_len() : target_max;
    target_min = (target_min <= 1) ? 1 : target_min;            // target_min mindestens 1

    // Alignment Grenzen: Query[1..m] mit Target[min_n, max_n]
    score_t tmp1, allmax = -INFTY;            // Grösster Score bisher
//    size_t arc_im_window = 0;

    matrix_entry del_query;               // Für affine Gapkosten, brauch ich eine zusätzliche Variable ...
    matrix_entry del_target[query_len+1]; // und ein Array (Ich muss wissen in welcher spalte (col) die deletion begann. Daher matrix_entry
    matrix_entry MAX;                     // maximaler Score für aktuelle Position

    // Ein Eintrag der Score Matrix besteht immer aus
    MAX.score = -INFTY;         //  dem Alignment score bisher
    MAX.start = -INFTY;         // und dem Anfangspunkt des Alignments (wird bis zum Alignmentende einfach durchgeschliffen

    if (show_alis){
        MAX.alignment = new char[max_ali_size];
        del_query.alignment = new char[max_ali_size];
        for (int i = 0; i <= query_len; i++)
            del_target[i].alignment = new char[max_ali_size];
    }

    // Pointer auf Liste von Arcs
    RightAdjList *arcsQuery  = NULL;
    RightAdjList *arcsTarget = NULL;

   // Rand initialisieren
    M(0, target_min - 1).score = 0;
    int gap_sum = scoring.gap_open_penalty();
    for (size_t2 i = 1; i <= query_len; i++) {   // local or semi-global alignment?
        gap_sum += scoring.gap_query(i);
//        M(i, target_min - 1).score = ( (locarna) ? (0) : -INFTY); // Kristin VersionLaut Kristin darf vom Query niemals was abgeschnitten werden. Dasher unendliche Strafe
        M(i, target_min - 1).score = ( (locarna) ? (0) : gap_sum); // /*scoring.gap_query(1, i)*/ ); // Linker Rand der Score Matrix bekommt Gap Strafen (s. semilokales Alignment)
        M(i, target_min - 1).start = target_min - 1;
        del_target[i].score = -INFTY;   // Ein Alignment beginnt nicht mit einer deletion
        del_target[i].start = -INFTY;

        if (show_alis)
            sprintf(M(i, target_min - 1).alignment, "%s%c-", M(i-1, target_min - 1).alignment, query[i]);
    }


    // Hier gehts los.  y.. positioen entlang des Target, x.. Position entlang der Query
    for (size_t2 y = target_min; y <= target_len; y++) { // Notiz an mich selbst, n == max_n

        ///////////// Only useless Output
        if (!NO_PROGRESS_OUTPUT)
            if ((score_t) (((y - 1 - target_min) / (float) (target_len - target_min))*100) < (score_t) (((y - target_min) / (float) (target_len - target_min))*100))
                std::cerr << "\r" << "  Processing " << target.seq_name() << ": " << (score_t) (((y - target_min) / (float) (target_len - target_min))*100) << "%" << std::flush; // Fortschritts Meldung der Konsole
        /////////////

        arcsTarget = &(target.right_adjlist(y)); // LISTE aller (target) Arcs deren RECHTES Ende bei y liegt
        // entferne alle target arcs deren linkes ende < target_min ist
	if ( (target_min > 1) && (arcsTarget->size() > 0) ){
		RightAdjList::iterator arc = arcsTarget->begin();
		while (arc != arcsTarget->end()){
			if (arc->left < target_min){
				arcsTarget->erase(arc);
				arc = arcsTarget->begin();
			}else
				arc++;
		}
	} // TODO: Testen ob ich die arcsTarget referenz nach dem loeschen von arcs nochmal neu laden muss, oder ob der weiterhin konsistent ist

//        arc_im_window += arcsTarget->size(); // SOVIELE ARCS KOMMEN NEU HINZU
        // Test ob unsere D Matrix genug Speicher hat
//while (arc_im_window >= D.cols()) D.redouble(); // Ggf. Speicher verdoppeln!


        ///  ES GEHT LOS
        M(0, y).score = 0; // Initialisation der aktuellen Spalte
        M(0, y).start = y+1;
                                  //M(0, y).alignment = "";  WEHE!!! strcpy nutzen!
        del_query.score = -INFTY;
	del_query.start = -INFTY;

        // x läuft von oben nach unten durch die aktuelle Spalte. (Also zeilenweise durch die matrix)
        for (size_t2 x = 1; x <= query_len; x++) {
            //del_query     = max1( M(x - 1, y    ) + scoring.gap_query(x,x), del_query + scoring.gap_extend_query(x,x));
            //del_target[x] = max1( M(x    , y - 1) + scoring.gap_target(y,y), del_target[x] + scoring.gap_extend_target(y,y));

            // Neuen Gap beginnen oder alten verlaengern
            if (M(x - 1, y).score + scoring.gap_query_open(x) > del_query.score){
                del_query.score = M(x - 1, y).score + scoring.gap_query_open(x); //gap_ext;  // neuer query gap auf
		del_query.start = M(x - 1, y).start;
                if (show_alis)
                    strcpy(del_query.alignment, M(x - 1, y).alignment);
            }
            del_query.score += scoring.gap_query(x); // gap_ext;     // alten query gap verlaengern

            if (M(x, y - 1).score + scoring.gap_open_penalty() > del_target[x].score){   // dasselbe fuer target
                del_target[x].score = M(x, y - 1).score + scoring.gap_open_penalty();
                del_target[x].start = M(x, y - 1).start;
                if (show_alis)
                        strcpy(del_target[x].alignment, M(x, y - 1).alignment);
            }
            del_target[x].score += scoring.gap_extend_penalty();

            //MAX = max2(M(x - 1, y - 1) + scoring.sigma(x, y), del_target[x], del_query);
              MAX.score = max2(  M(x - 1, y - 1).score + scoring.sigma(x, y),        // (mis)match
                                del_target[x].score,                                // insert
                                del_query.score);                                   // deletion

             if (MAX.score == del_target[x].score)              // deletion
                 MAX.start = del_target[x].start;
             else if (MAX.score == del_query.score)             // insert
                 MAX.start = del_query.start;
             else
                 MAX.start = M(x - 1, y - 1).start;             // (mis)match

             if (show_alis){
                 int l = strlen(del_query.alignment);
                 del_query.alignment[l]   = query[x];
                 del_query.alignment[l+1] = '-';
                 del_query.alignment[l+2] = '\0';

                 l = strlen(del_target[x].alignment);
                 del_target[x].alignment[l]   = '-';
                 del_target[x].alignment[l+1] = target[y];
                 del_target[x].alignment[l+2] = '\0';

                if (MAX.score == del_target[x].score)              // deletion
                    strcpy(MAX.alignment, del_target[x].alignment);
                else if (MAX.score == del_query.score)             // insert
                    strcpy(MAX.alignment, del_query.alignment);
                else
                    sprintf(MAX.alignment, "%s%c%c", M(x - 1, y - 1).alignment, query[x], target[y]);
             }

//            MAX.score = max2(M(x-1, y-1).score + scoring.sigma(x,y), M(x-1,y).score + gap_ext, M(x,y-1).score + gap_ext);

            // Check if we have arcs here
            if (arcsTarget->size() > 0) {
                // We need to compute an global(!) alignment between the longest arc in target against the longest arc in query!
                // All alignment scores between shorter arcs, can be found in the same matrix.
                // We compute a suffix-alignment!
                arcsQuery = &query.right_adjlist(x); // Liste alle (query) arcs deren RECHTES Ende bei x liegt
                bool gefunden = false;

                /* es passieren zwei Dinge. Solange keine gueltige Arc-Arc Kombination gefunden wurde, laufen wir durch die moeglichen arcs
                 * vom groessten zu immer kleineren. Sobald gefunden==true wird EINMAL das globale Alignment zwischen beiden berechnet.
                 * Alle weiteren (kleineren) Arcs werden direkt aus der globalen Scoring matrix gelesen */
                for (RightAdjList::const_iterator q_arc = arcsQuery->begin(); q_arc != arcsQuery->end(); q_arc++)
                    for (RightAdjList::const_iterator t_arc = arcsTarget->begin(); t_arc != arcsTarget->end(); t_arc++){
                        if ( (!gefunden) && lohnt_sich(*q_arc, *t_arc) ){ // suchen ob die arc kombination gut ist
                            gefunden = true;
                            global_align(q_arc->left + 1, q_arc->right - 1,  // globales alignment ausrechnen
                                         t_arc->left + 1, t_arc->right - 1);
                           } // TODO: DARF ICH WIRKLICH ALLE WEITEREN KOMBOS DIREKT AUSLESEN ODER NICHT?!

                        // TODO: MOEGLICHE FEHLERQUELLE NOCHMAL UEBERDENKEN!!!!
                        if (lohnt_sich(*q_arc, *t_arc)){
                            // Alle restlichen Arc-Kombinationen lassen sich direkt aus der Temp Matrix lesen
                            // Speichern in der D-Matrix
                            D(q_arc->idx, t_arc->idx).score = tmp(q_arc->left + 1, t_arc->left + 1).score + scoring.arcmatch((*q_arc), (*t_arc));

                            if (show_alis){
                                if (D(q_arc->idx, t_arc->idx).alignment == NULL)
                                        D(q_arc->idx, t_arc->idx).alignment = new char[max_ali_size];
                                strcpy(D(q_arc->idx, t_arc->idx).alignment, tmp(q_arc->left + 1, t_arc->left + 1).alignment);
                            }
                            // Und nebenbei die berechnung des aktuellen Score  --> DER LETZTE SANKOFF SCHRITT
                            tmp1 = M(q_arc->left - 1, t_arc->left - 1).score + D(q_arc->idx, t_arc->idx).score;
                            if (tmp1 > MAX.score){
                                MAX.score = tmp1;
                                MAX.start = M(q_arc->left - 1, t_arc->left - 1).start;

                                if (show_alis) // ich codiere die öffnenden und schliessenden Gaps indem ich das Query_character um +41 '(' oder +82 ')'  verschiebe!
                                   // sprintf(MAX.alignment, "%s(%c%c%s%c%c)", M(q_arc->left - 1, t_arc->left - 1).alignment, query[q_arc->left], target[t_arc->left],
                                   //        D(q_arc->idx, t_arc->idx).alignment, query[q_arc->right], target[t_arc->right]);    // Mentale Notiz, query[left] target[left] sind auf jeden Fall aligniert! Da gibts einfach kein Gap!
                                    sprintf(MAX.alignment, "%s%c%c%s%c%c", M(q_arc->left - 1, t_arc->left - 1).alignment, query[q_arc->left]+41, target[t_arc->left],
                                           D(q_arc->idx, t_arc->idx).alignment, query[q_arc->right]+82, target[t_arc->right]);    // Mentale Notiz, query[left] target[left] sind auf jeden Fall aligniert! Da gibts einfach kein Gap!
                            }
                        }
                }
            }


	    if ((locarna) && (MAX.score < 0)) {
		MAX.score = 0;
		MAX.start = y;
	    }

            // FINALLY M(x, y) = MAX;
            M(x, y).score = MAX.score;
            M(x, y).start = MAX.start;
            if (show_alis)
                strcpy(M(x, y).alignment, MAX.alignment);           // Wichtig zu kopieren!
        }


        if (y >= target_min + MAX_SPAN){
            //arc_im_window -= target.right_adjlist(y - MAX_SPAN).size(); // Soviele ARCS fliegen raus
            arcsTarget = &target.right_adjlist(y - MAX_SPAN);
            for (RightAdjList::const_iterator t_arc = arcsTarget->begin(); t_arc != arcsTarget->end(); t_arc++)
                for (RightAdjList::const_iterator q_arc = arcsQuery->begin(); q_arc != arcsQuery->end(); q_arc++)
                        if (lohnt_sich(*q_arc, *t_arc))
                            D.erase(q_arc->idx, t_arc->idx);
        }

        if ((y-target_min) >= query_len){       // Nur wenn das Alignment ueberhaupt schon lang genug ist, wird es gespeichert!
            if (locarna){
                int mx = 1;
                for (int h = 1; h <= query_len; h++)
                    if (M(h, y).score > M(mx, y).score)
                        mx = h;
                if (M(mx, y).score > allmax){
                    allmax = M(mx, y).score; // Aktualisieren des Maximales Score
                }
                store_score(M(mx, y).start, y, M(mx, y).score, M(mx, y).alignment);}
            else{
                if (M(query_len, y).score > allmax)
                    allmax = M(query_len, y).score;
                store_score(M(query_len, y).start, y, M(query_len, y).score, M(query_len, y).alignment);
            }
         }
    }


   if (growing_score)
                local_max_list.push(last_score);

    if (!NO_PROGRESS_OUTPUT)
        std::cerr << "\r                                                                                  \r" << std::flush;

    histogram_file.close();
    if (show_alis){
        delete[] MAX.alignment;
        delete[] del_query.alignment;
        for (int i = 0; i <= query_len; i++)
            delete[] del_target[i].alignment;

    }
    return allmax;
}
