#ifndef aligner_hh
#define aligner_hh


#include "matrices.hh"
#include "ScoringModell.hh"
#include <iostream>
#include <fstream>
#include <string>
#include <list>
#include <map>
#include "Query.hh"
#include "Target.hh"
#include "base.hh"
#include <vector>

void pretty_print_komprimiert(const std::string ali, const size_t2 start=-1, const size_t2 end=-1);

class Aligner {
private:
	result_type last_score;
	bool growing_score;
        bool locarna;
        LeftAdjList *tmp_arcsQuery;
	LeftAdjList *tmp_arcsTarget;
        char *alignment_strings;            // Hier kommen ALLE Alignments (string) rein.
        char *tmp_alignment_strings;        // Hier kommen alle temporaerent Alignments rein (der tmp matrix).
        const bool show_alis;               // Sollen die Alignments komplett ausgegeben werden, oder reichen Anfangs- und Endpunkte?
        const size_t max_ali_size;             // Maximale Laenge eines Alignment Strings
        const int query_len;                // zum schnellen Zugriff auf die statische Query Sequenz Laenge

        matrix_entry tmp_del_query, tmp_MAX;    // Damit ich bei global align nicht dauernd neu speicher allokieren muss
        matrix_entry *tmp_del_target;
protected:
    Query &query;
    Target &target;

    std::ofstream histogram_file;
    PriorityQueue k_best_list;          // stores the k-best alignment scores (and positions)
    PriorityQueue local_max_list;       // stores enough local maximas for later filtering
    ScoringModell &scoring;
    Rotation_matrix<matrix_entry> M;    // Scoring Matrix für lokale Alignments
    //Rotation_matrix<matrix_entry> D;    // Speichert nur den Alignment Score eines arcumschlossenen Bereiches


    D_Matrix D;

    OffsetMatrix<matrix_entry> tmp;     // temp matrix um D-Einträge auszurechnen

    //Rotation_matrix<matrix_entry> tmp;     // temp matrix um D-Einträge auszurechnen

    score_t MAX_SPAN;

    void global_align(size_t2 i, size_t2 j, size_t2 k, size_t2 l);
    void store_score(const size_t2 start, const size_t2 end, const int score, char * ali=NULL);

    void init_matrix(size_t2 col){
    	size_t2 m = query.seq_len();
    	for (size_t2 i = 1; i<=m; i++){
    		M(i,col).score = scoring.gap_query(1,i);
                M(i,col).start = col;
                M(i,col).alignment = NULL;
        }
    	M(0,col).score = 0;
        M(0,col).start = col;
        M(0,col).alignment = NULL;
    }

    bool lohnt_sich(const Arc &q, const Arc &t);

    void find_kbest();

public:

    bool NO_PROGRESS_OUTPUT;
    //! construct with sequences and corresponding sets of basepairs
    Aligner(ScoringModell &scoring, Query &_query, Target &_target, std::string output_file, bool show_alis);

    ~Aligner();

    score_t local_align(size_t2 min_n = 1, size_t2 max_n = -1);
	void free_memory(size_t2 i);

    PriorityQueue & get_k_best();
    PriorityQueue & get_lm_list();

    // Locales statt semiglobales Alignment
    void enable_query_local_alignment(void);

    //! compute the alignment score
    score_t search();
};

#endif




