/* evalueCALC.c
 (C) Jana Hertel
 */
#include "evalueCALC.h"

/* -------------------------------------------------------------------------- */

void gotoh_getFitRange(float **logdata, int n, int *s0, int *cut) {        
  float maxY=0;
  int i, maxI, maxX;  
  for(i=0; i<n; i++) {      
    if(logdata[1][i] > maxY) {
      maxY = logdata[1][i];
      maxI = i;
    }
  }
  maxX = logdata[0][maxI];      // == Der am haeufigsten vorkommende Score
 
  float halfY = 0.5 * maxY;     // max_counts / 2
  int halfI;                    // position of halfY

  for(i=maxI; i<n; i++) {
    if(logdata[1][i] < halfY) break;
    halfI = i;
  }
  halfY = logdata[1][halfI];   /* true datapoint */ // Next datapoint <= max_count / 2  
  float range = 0.125, lcutY, rcutY;
  int lcutI, rcutI;

  lcutY = halfY + range*maxY;   // linker Cut = max_count / 2 + 1/8 * max_count
  lcutI = maxI; // PATCH
  for(i=maxI; i<halfI; i++) {
    if(logdata[1][i] < lcutY) break;
    lcutI = i;
  }
  lcutY = logdata[1][lcutI];

  rcutY = halfY - range*maxY;
  rcutI = halfI+1; // PATCH
  for(i=halfI+1; i<n; i++) {
    if(logdata[1][i] < rcutY)
        break;
    rcutI = i;
  }
  rcutY = logdata[1][rcutI];

  int lm_size = rcutI - lcutI + 1;
  double *x=NULL, *y=NULL;

  x = (double *) calloc(lm_size + 1, sizeof(double));
  y = (double *) calloc(lm_size + 1, sizeof(double));
  for(i=0; i<lm_size; i++) {
    y[i] = logdata[1][lcutI+i];
    x[i] = logdata[0][lcutI+i];
  }


   fprintf(stdout, "#Tangent fitting\n");
   fprintf(stdout, "#max:     %d %f\n", maxX, maxY);
   fprintf(stdout, "#1/2 max: %d %f\n", (int)(logdata[0][halfI]+0.5), halfY);
   fprintf(stdout, "#lcut:    %d %f\n", (int)(logdata[0][lcutI]+0.5), lcutY);
   fprintf(stdout, "#rcut:    %d %f\n", (int)(logdata[0][rcutI]+0.5), rcutY);
   fprintf(stdout, "#------------\n");

  double A=0, B=0, C=0, D=0;
  int S;

  for(i=0; i<lm_size; i++) {
    A += x[i];
    B += y[i];
    C += x[i]*x[i];
    D += x[i]*y[i];
  }
  S = lm_size;

  // fprintf(stderr, "A: %f\n", A);
//   fprintf(stderr, "B: %f\n", B);
//   fprintf(stderr, "C: %f\n", C);
//   fprintf(stderr, "D: %f\n", D);

  double m, N;

  m = (S*D - A*B) / (S*C - A*A);
  N = (D - m*C) / A;

   fprintf(stdout, "#y=%fx + %f\n", m, N);

  int ci;

  ci = (int) ((-N)/m + 0.5);
  for(i=halfI; i<n; i++)
    if(logdata[0][i] > ci) break;

  *cut = i - 1;

  int s0i;

  s0i = (int) ((logdata[0][halfI] - maxX + 1.5) / 2 + maxX);
  for(i=maxI; i<halfI; i++)
    if(logdata[0][i] > s0i) break;

  *s0 = i - 1;

  free(x); free(y);
  return;

}


/* -------------------------------------------------------------------------- */

void gotoh_calcEvalue(double *evalues, int *k_best, int size, float **logdata, int s0, int cut) {
  int i, n;

  n = cut - s0 + 1;

  double * x = (double *) calloc(n, sizeof(double));
  double * y = (double *) calloc(n, sizeof(double));
  for(i=0; i<n; i++) {
    x[i] = logdata[0][s0+i];
    y[i] = logdata[1][s0+i];
  }

  double A=0, B=0, C=0, D=0, E=0, F=0, G=0, H=0;
  int S;
  for(i=0; i<n; i++) {
    A += x[i];
    B += log(x[i]);
    C += x[i]*log(x[i]);
    D += x[i]*x[i];
    E += log(x[i])*log(x[i]);
    F += x[i]*y[i];
    G += y[i]*log(x[i]);
    H += y[i];
  }
  S = n;

  free(x); free(y);

  double a0, a1, a2;

  a2 = ((A*H-S*F)*(A*B-S*C) - (B*H-S*G)*(A*A-S*D)) / ((A*B-S*C)*(A*B-S*C) - (A*A-S*D)*(B*B-S*E));
  a1 = (A*H - S*F - (A*B-S*C)*a2) / (A*A-S*D);
  a0 = (H - B*a2 - A*a1) / S;

  fprintf(stdout, "#fitted polynomial: y = a2 * log(x) + a1 * x + a0\n"); 
  fprintf(stdout, "#a0: %f\n", a0);
  fprintf(stdout, "#a1: %f\n", a1);
  fprintf(stdout, "#a2: %f\n", a2);

  double k = 1/exp(gotoh_psi(logdata[0][cut], a0, a1, a2));

  for(i=0; i<size; i++)
    evalues[i] = k*exp(gotoh_psi(k_best[i], a0, a1, a2));
}

/* -------------------------------------------------------------------------- */

float gotoh_psi(int score, double a0, double a1, double a2) {
    return a0 + a1*(score + 1) + a2*log(score + 1);
}

