/* evalueCALC.h
 (C) Jana Hertel
 */

#ifndef eCALC_H
#define eCALC_H

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>


/**
   \brief Determine the fit range.

   \param data Distribution of local maxima. 2 Dimensions: 0->xvals, 1->yvals
   \param n Size of Distribution.
   \param s0 Left border.
   \param cut Right border.
   \param slope Slope of tangente.
   \param intercept Intercept of tangente.

   Determine the left and right borders of the tail of the data
   that should be used for fitting.

 */
void gotoh_getFitRange(float **data, int n, int *s0, int *cut);

/**
   \brief Fit the psi funktion.

   \param L List of local maxima. Complete backtraced alignments.
   \parma sizeL Size of List L.
   \param logdata Distribution of local maxima. 2 Dimensions: 0->xvals, 1->yvals
   \param sizeD Size of Distribution.
   \param s0 Left border.
   \param cut Right border.
   \param maxPossScore Highest possible Alignment score.
 */
void gotoh_calcEvalue(double* evalues, int* kbest, int size, float **logdata, int s0, int cut);
void gotoh_calcEvalue2(double* evalues, int* kbest, int size, float **logdata, int s0, int cut);

float gotoh_psi(int score, double a0, double a1, double a2);
float gotoh_psi2(int score, double a0, double a1, double a2);


#endif

/* End of file */
