#ifndef PARAMS_HH
#define PARAMS_HH

#include <iostream>
#include "base.hh"
#include <cmath>
/*
  Parameter for Aligner
*/

class AlignerParams {
public:
    const int match_score;
    const int mismatch_score;
    const int gap_open_score;
    const int gap_extend_score;
    const int tau_factor; // contribution of sequence similarity in an arc match (in percent)
    const int max_diff;
    const int max_diff_am;
    const int max_span;  // Longest arc span

    AlignerParams(int _match_score,int _mismatch_score,int _gopen_score, int _gextend_score, int _tau_factor, int _max_diff, int _max_diff_am, int _max_span):
	match_score(_match_score),
	mismatch_score(_mismatch_score),
	gap_open_score(_gopen_score),
	gap_extend_score(_gextend_score),
	tau_factor(_tau_factor),
	max_diff(_max_diff),
	max_diff_am(_max_diff_am),
	max_span(_max_span){}
};


/*!
  parameters for the basepairs class

  contains the necessary information for determining the
  significant basepairs and computing their weights
*/
class BasePairsParams {
public:
    const int structure_weight; //!< weight for balancing structure contribution to sequence contribution in the score

    double prob_exp_f(size_t2 seqlen) const {return 1.0/(2.0*seqlen);} //!<expected probability of a base pair (null-model)

    score_t probToWeight(const double p, size_t2 seqlen) const {
	//double pe=prob_exp_f(seqlen);
	double pe=0.3; /* TODO! HARD CODED AD HOC VALUE */
	return (score_t)round(structure_weight * (1-log(p)/log(pe)));
    }

    const double min_prob; //!< only pairs with a probability of at least min_prob are taken into account
    
    BasePairsParams(int _structure_weight, double _min_prob)
      :
	structure_weight(_structure_weight),
	min_prob(_min_prob)
    {}
};

#endif // PARAMS_HH
